#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


import numpy as np
import itertools
import contextlib
from wield.bunch import Bunch
from wield.bunch.depbunch import (
    NOARG,
    depB_property,
)


from ..external import scipy_optimize
from .. import TFmath
from .. import representations
from ..utilities import ensure_aid

from .data_filtfit_base import DataFiltFitZ, DataFiltFitBase, DataFiltFitSf


class MultiReprFilterBase(DataFiltFitBase):
    # TODO add a "use_phase" option (defaulting to True)
    phase_missing = False
    rcond_cov = 1e-5
    root_constraint = representations.root_constraints.mirror_real
    RBalgo = representations.RBAlgorithms(constraint_standard=root_constraint)

    def __build__(
        self,
        _args=None,
        parent=None,
        gain_coding=None,
        delay_coding=None,
        num_codings=None,
        den_codings=None,
        parameters=None,
        poles_overlay=None,
        zeros_overlay=None,
        residuals_type=None,
        check_sign=None,
        codings_ignore=None,
        residuals_log_im_scale=None,
        distance_limit_auto=None,
        distance_limit_scale=None,
        delay_s_min=None,
        delay_s_max=None,
        coding_map=None,
        h_infinity=None,
        h_infinity_deweight=None,
        max_BW_Hz=None,
        copy=None,
        **kwargs
    ):
        if _args is not None:
            raise RuntimeError("Only keyword arguments allowed")
        if copy is not None:
            parent = copy
        super(MultiReprFilterBase, self).__build__(parent=parent, copy=copy, **kwargs)
        if parent is not None and isinstance(parent, MultiReprFilterBase):
            if parameters is None:
                parameters = parent.parameters
            if gain_coding is None:
                gain_coding = parent.gain_coding
            if delay_coding is None:
                delay_coding = parent.delay_coding
            # TODO, use the copier
            if num_codings is None:
                num_codings = parent.num_codings
            # TODO, use the copier
            if den_codings is None:
                den_codings = parent.den_codings
            if poles_overlay is None:
                poles_overlay = parent.poles_overlay
            if zeros_overlay is None:
                zeros_overlay = parent.zeros_overlay
            if residuals_type is None:
                residuals_type = parent.residuals_type
            if check_sign is None:
                check_sign = parent.check_sign
            if codings_ignore is None:
                arg_codings_ignore = parent.codings_ignore
            if residuals_log_im_scale is None:
                residuals_log_im_scale = parent.residuals_log_im_scale
            if distance_limit_auto is None:
                distance_limit_auto = parent.distance_limit_auto
            if distance_limit_scale is None:
                distance_limit_scale = parent.distance_limit_scale
            if delay_s_min is None:
                delay_s_min = parent.delay_s_min
            if delay_s_max is None:
                delay_s_max = parent.delay_s_max
            if coding_map is None:
                coding_map = parent.coding_map
            if h_infinity is None:
                h_infinity = parent.h_infinity
            if h_infinity_deweight is None:
                h_infinity_deweight = parent.h_infinity_deweight
            if max_BW_Hz is None:
                max_BW_Hz = parent.max_BW_Hz
        else:
            if gain_coding is None:
                raise RuntimeError("Must specify gain_coding")
            if num_codings is None:
                raise RuntimeError("Must specify num_codings")
            if den_codings is None:
                raise RuntimeError("Must specify den_codings")
            if poles_overlay is None:
                poles_overlay = ()
            if zeros_overlay is None:
                zeros_overlay = ()
            if residuals_type is None:
                residuals_type = "log"
            if check_sign is None:
                check_sign = False
            if codings_ignore is None:
                arg_codings_ignore = set()
            if residuals_log_im_scale is None:
                residuals_log_im_scale = 1
            if distance_limit_auto is None:
                distance_limit_auto = 0
            if distance_limit_scale is None:
                distance_limit_scale = 1
        arg_gain_coding = gain_coding
        arg_delay_coding = delay_coding
        arg_den_codings = den_codings
        arg_num_codings = num_codings

        self.max_BW_Hz = max_BW_Hz
        # must adjust this whenever something happens to the codings
        self.codings_revision = 0

        self.coding_map = coding_map
        self.h_infinity = h_infinity
        self.h_infinity_deweight = h_infinity_deweight

        # can be 0: never, 1: on call, or 2: on all parameter update calls (during optimization)
        self.distance_limit_auto = distance_limit_auto
        self.distance_limit_scale = distance_limit_scale

        self.gain_coding = arg_gain_coding
        self.num_codings = arg_num_codings
        self.den_codings = arg_den_codings

        if arg_delay_coding is not None:
            self.delay_coding = arg_delay_coding
        if delay_s_min is not None:
            self.delay_s_min = delay_s_min
        if delay_s_max is not None:
            self.delay_s_max = delay_s_max

        self.poles_overlay = poles_overlay
        self.zeros_overlay = zeros_overlay
        self.residuals_type = residuals_type
        self.check_sign = check_sign
        self.residuals_log_im_scale = residuals_log_im_scale
        # assign after the setter is built
        self.codings_ignore = arg_codings_ignore
        self.auto_jacobian = False

        # the parameters are mutable objects that need copy rules
        @self.deco_copier
        def gain_coding(self, val):
            return val.clone(self)

        @self.deco_copier
        def num_codings(self, val):
            return [c.clone(self) for c in val]

        @self.deco_copier
        def den_codings(self, val):
            return [c.clone(self) for c in val]

    @depB_property
    def h_infinity(self, val=NOARG):
        if val is NOARG or val is None:
            val = 0
        if callable(val):
            val = val(len(self.F_Hz))
        val = np.asarray(val)
        return val

    @depB_property
    def h_infinity_deweight(self, val=NOARG):
        if val is NOARG or val is None:
            val = 0.1

        val = float(val)
        return val

    @depB_property(autodeps=False)
    def residuals_log_im_scale(self, val=1):
        # need to set without putting a dep on this value (or it will reset)
        self.codings_revision = 1 + self.get_raw("codings_revision")
        return np.asarray(val)

    @depB_property
    def delay_s_min(self, val=NOARG):
        if val is NOARG:
            # default min delay is 22.5deg loss by the last frequency point
            # val = -(np.pi/8) / self.F_max_Hz
            val = 0
        else:
            # need to set without putting a dep on this value (or it will reset)
            self.codings_revision = 1 + self.get_raw("codings_revision")
        return val

    @depB_property
    def delay_s_max(self, val=NOARG):
        if val is NOARG:
            # default max delay is 10deg loss by the last frequency point
            val = (np.pi / 16) / self.F_max_Hz
        else:
            # need to set without putting a dep on this value (or it will reset)
            self.codings_revision = 1 + self.get_raw("codings_revision")
        return val

    @depB_property
    def max_BW_Hz(self, val=NOARG):
        if val is NOARG or val is None:
            val = 3 * self.F_max_Hz
        return val

    @depB_property(autodeps=False)
    def distance_limit_auto(self, val):
        assert val in [0, 1, 2]
        # need to set without putting a dep on this value (or it will reset)
        self.codings_revision = 1 + self.get_raw("codings_revision")
        return val

    @depB_property
    def zeros_overlay(self, val=NOARG):
        if val is NOARG:
            return self.RBalgo.expect([])
        else:
            rB = self.RBalgo.expect(val)
            return rB

    @depB_property
    def poles_overlay(self, val=NOARG):
        if val is NOARG:
            return self.RBalgo.expect([])
        else:
            rB = self.RBalgo.expect(val)
            return rB

    @depB_property
    def _extra_xfer_fit(self):
        # TODO must add pole-zero rephasing for Z filters
        xfer_p, lnG_p = self.poles_overlay.val_lnG(self.Xex_grid)
        xfer, lnG = self.zeros_overlay.val_lnG(self.Xex_grid, h=1 / xfer_p, lnG=-lnG_p)
        return xfer * np.exp(lnG)

    @depB_property
    def data_no_overlay(self):
        return self.data / self._extra_xfer_fit

    @depB_property
    def xfer_fit_no_overlay(self):
        return self.xfer_fit / self._extra_xfer_fit

    @depB_property
    def data_effective(self):
        return self.data

    @depB_property(autodeps=False)
    def gain_coding(self, val=NOARG):
        if val is NOARG:
            raise RuntimeError("Currently can't generate a gain coding")
        else:
            # need to set without putting a dep on this value (or it will reset)
            self.codings_revision = 1 + self.get_raw("codings_revision")
            return val.clone(self)

    @depB_property(autodeps=False)
    def delay_coding(self, val=NOARG):
        if val is NOARG:
            return self.coding_map.delay(self)
        else:
            return val.clone(self)

    @depB_property(autodeps=False)
    def num_codings(self, val):
        self.codings_revision = self.codings_revision + 1
        return [c.clone(self) for c in val]

    @depB_property(autodeps=False)
    def den_codings(self, val):
        self.codings_revision = self.codings_revision + 1
        return [c.clone(self) for c in val]

    @depB_property
    def root_codings_set(self):
        return set(self.num_codings + self.den_codings)

    @depB_property
    def codings_ignore(self, val):
        return val

    @depB_property
    def codings_ignore_mapped(self):
        self.codings_revision = self.codings_revision + 1
        val = self.codings_ignore
        if val is None:
            val = set()
        else:
            coding_id_map = {self.gain_coding.coding_id: self.gain_coding}
            if self.delay_coding is not None:
                coding_id_map[self.delay_coding.coding_id] = self.delay_coding
            for coding in self.num_codings:
                coding_id_map[coding.coding_id] = coding
            for coding in self.den_codings:
                coding_id_map[coding.coding_id] = coding

            codings_ignore_mapped = set()
            for coding in val:
                mcoding = coding_id_map.get(coding.coding_id, None)
                if mcoding is not None:
                    codings_ignore_mapped.add(mcoding)
            val = codings_ignore_mapped
        return val

    @depB_property
    def coding_lists(self):
        num_active = []
        num_ignore = []
        den_active = []
        den_ignore = []
        if self.gain_coding not in self.codings_ignore_mapped:
            num_active.append(self.gain_coding)
        else:
            num_ignore.append(self.gain_coding)

        if (
            self.delay_coding is not None
            and self.delay_coding not in self.codings_ignore_mapped
        ):
            num_active.append(self.delay_coding)
        else:
            num_ignore.append(self.delay_coding)

        for idx, coding in enumerate(self.num_codings):
            if coding not in self.codings_ignore_mapped:
                num_active.append(coding)
            else:
                num_ignore.append(coding)
        for idx, coding in enumerate(self.den_codings):
            if coding not in self.codings_ignore_mapped:
                den_active.append(coding)
            else:
                den_ignore.append(coding)
        return Bunch(
            num_active=num_active,
            num_ignore=num_ignore,
            den_active=den_active,
            den_ignore=den_ignore,
        )

    @depB_property
    def xfer_fit(self, val=NOARG):
        if val is NOARG:
            # access just for the dependency
            # print("NEW_XFER")
            self.codings_revision
            if self.auto_jacobian:
                xfer, jac = self.transfer_jacobian_generate()
                self.residuals_jacobian = jac
                # TODO hack
                self.codings_revision
                return xfer
            else:
                xfer = self.transfer_generate()
                # TODO hack
                self.codings_revision
                return xfer
        else:
            # access just for the dependency
            self.codings_revision
            return val

    @depB_property
    def xfer_inactive(self):
        # access just for the dependency
        self.codings_revision
        return self.transfer_generate(generate_inactive=True)

    @depB_property
    def residuals_jacobian(self, val=NOARG):
        if val is NOARG:
            # access just for the dependency
            # print("NEW_JAC")
            self.codings_revision
            xfer, jac = self.transfer_jacobian_generate()
            # TODO can do this assignment, but need to set the dependencies with it
            self.xfer_fit = xfer
            return np.asarray(jac)
        else:
            # access just for the dependency
            self.codings_revision
            return val

    @depB_property
    def parameters(self, params=NOARG):
        if params is NOARG:
            params = []
            N_params = 0
            for coding in self.coding_lists.num_active:
                N_params += coding.N_parameters
                c_params = np.asarray(coding.reduce())
                # TODO, move these checks into the parameters
                if any(c_params.imag != 0):
                    raise RuntimeError(
                        "Coding not using real parameters {0}".format(coding)
                    )
                params.extend(c_params.real)
            for coding in self.coding_lists.den_active:
                N_params += coding.N_parameters
                c_params = np.asarray(coding.reduce())
                # TODO, move these checks into the parameters
                if any(c_params.imag != 0):
                    raise RuntimeError(
                        "Coding not using real parameters {0}".format(coding)
                    )
                params.extend(c_params.real)
            # add the dependency
            self.dependencies("codings_revision")
            return np.asarray(params, dtype=float)
        else:
            prev_params = self.get_raw("parameters", None)
            if (prev_params is not None) and np.all(prev_params == params):
                # print("SAME")
                return params

            # access for the dependency, and update the value to clear other dependent values
            self.codings_revision = self.codings_revision + 1

            N_current = 0
            for coding in self.coding_lists.num_active:
                num_next = coding.N_parameters
                coding.update(*params[N_current : N_current + num_next])
                N_current += num_next
            for coding in self.coding_lists.den_active:
                num_next = coding.N_parameters
                coding.update(*params[N_current : N_current + num_next])
                N_current += num_next
            assert N_current == len(params)
            return params

    @depB_property
    def coding_maps(self):
        # TODO, use coding_lists, but have to handle gain_coding
        mapping = []
        locations_map = {}
        parameters_map = {}
        N_current = 0
        if self.gain_coding not in self.codings_ignore_mapped:
            mapping.extend(self.gain_coding.N_parameters * [self.gain_coding])
            num_next = self.gain_coding.N_parameters
            parameters_map[self.gain_coding] = N_current
            N_current += num_next
            locations_map[self.gain_coding] = "gain_coding"

        if (
            self.delay_coding is not None
            and self.delay_coding not in self.codings_ignore_mapped
        ):
            mapping.extend(self.delay_coding.N_parameters * [self.delay_coding])
            num_next = self.delay_coding.N_parameters
            parameters_map[self.delay_coding] = N_current
            N_current += num_next
            locations_map[self.delay_coding] = "delay_coding"

        num_idx = N_current
        for idx, coding in enumerate(self.num_codings):
            if coding in self.codings_ignore_mapped:
                continue
            num_next = coding.N_parameters
            parameters_map[coding] = N_current
            N_current += num_next
            mapping.extend(coding.N_parameters * [coding])
            locations_map[coding] = ("num_codings", idx)
        den_idx = N_current
        for idx, coding in enumerate(self.den_codings):
            if coding in self.codings_ignore_mapped:
                continue
            num_next = coding.N_parameters
            parameters_map[coding] = N_current
            N_current += num_next
            mapping.extend(coding.N_parameters * [coding])
            locations_map[coding] = ("den_codings", idx)
        # access for the dependency
        self.codings_revision
        return Bunch(
            p2c=mapping,
            c2loc=locations_map,
            c2p=parameters_map,
            num_idx=num_idx,
            den_idx=den_idx,
        )

    @depB_property
    def zeros(self, val=NOARG):
        if val is not NOARG:
            raise RuntimeError("Zeros not Directly Settable")
        # access just for the dependency
        self.codings_revision
        roots_c = []
        for coding in self.num_codings:
            c = coding.roots_c()
            assert np.all(np.imag(c) >= 0)
            roots_c.extend(c)
        return representations.RootBunch(
            c=roots_c,
            r=[r for coding in self.num_codings for r in coding.roots_r()],
            constraint=representations.root_constraints.mirror_real,
        )

    @depB_property
    def poles(self, val=NOARG):
        if val is not NOARG:
            raise RuntimeError("Poles not Directly Settable")
        # access just for the dependency
        self.codings_revision
        roots_c = []
        for coding in self.den_codings:
            c = coding.roots_c()
            assert np.all(np.imag(c) >= 0)
            roots_c.extend(c)
        return representations.RootBunch(
            c=roots_c,
            r=[r for coding in self.den_codings for r in coding.roots_r()],
            constraint=representations.root_constraints.mirror_real,
        )

    @depB_property
    def _gain_effect(self):
        # TODO, add gain properties to not need the raw parameter
        self.codings_revision
        gain_effect = 1
        for coding in self.num_codings:
            gain_effect *= coding.gain_effect
        for coding in self.den_codings:
            gain_effect /= coding.gain_effect
        return gain_effect

    @depB_property
    def gain(self, gain=NOARG):
        if gain is NOARG:
            # TODO, add gain properties to not need the raw parameter
            self.codings_revision
            return self.gain_coding.gain * self._gain_effect
        else:
            # TODO, add gain properties to not need the raw parameter
            gain_effect = self._gain_effect
            self.gain_coding.gain = gain / gain_effect
            self.codings_revision = self.codings_revision + 1
            return gain

    @depB_property
    def delay_s(self, val=NOARG):
        if val is NOARG:
            # TODO, add gain properties to not need the raw parameter
            self.codings_revision
            if self.delay_coding is None:
                return 0
            else:
                return self.delay_coding.delay_s
        else:
            self.codings_revision = self.codings_revision + 1
            self.delay_coding.delay_s = val
        return val

    @depB_property
    def residuals(self):
        self.codings_revision
        retB = Bunch()
        R = self.xfer_fit / self.data
        retB.resP = self.W * (R - 1)
        retB.resZ = self.W * (1 / R - 1)
        retB.resD = self.W * (R - 1 / R)

        R_abs_sq = TFmath.abs_sq(R)
        R_abs = R_abs_sq ** 0.5
        log_re = self.W * np.log(R_abs)
        log_im = self.W * R.imag / R_abs
        retB.resL = log_re + 1j * log_im * self.residuals_log_im_scale
        return retB

    @depB_property
    def residuals_preferred(self):
        self.codings_revision
        return self.residuals_compute(self.data, self.xfer_fit, self.W)

    @depB_property
    def residuals_average(self):
        if np.any(self.residuals_log_im_scale >= 1e-12):
            return np.sum(TFmath.abs_sq(self.residuals_preferred)) / (
                2 * len(self.data) - len(self.parameters)
            )
        else:
            return np.sum(TFmath.abs_sq(self.residuals_preferred)) / (
                len(self.data) - len(self.parameters)
            )


    @depB_property
    def residuals_sq(self):
        return TFmath.abs_sq(self.residuals_preferred)

    @depB_property
    def residuals_max(self):
        # includes adjustment for parameter number. Not a real thing statistically...
        return np.max(TFmath.abs_sq(self.residuals_preferred))

    @depB_property
    def residuals_med(self):
        # includes adjustment for parameter number. Not a real thing statistically...
        return np.median(TFmath.abs_sq(self.residuals_preferred))

    @depB_property
    def fisher_matrix(self):
        # TODO, compute this using QR decomposition first
        # followed perhaps by SVD
        pjac = self.residuals_jacobian
        phess = np.dot(pjac, pjac.conjugate().T).real
        return phess

    @depB_property
    def fisher_matrix_rel(self):
        # TODO, compute this using QR decomposition first
        # followed perhaps by SVD
        mat = np.copy(self.fisher_matrix)
        d = mat.diagonal() ** 0.5

        # recondition to prevent NANs
        d[d < 1e-15] = 1

        mat = mat / d.reshape(-1, 1)
        mat = mat / d.reshape(1, -1)
        return mat

    @depB_property
    def cov_matrix(self):
        # cov = np.linalg.pinv(self.fisher_matrix, rcond=1e-10)
        # uses the reconditioned fisher_matrix. This allows it to keep even the very badly behaved
        # values
        d = self.fisher_matrix.diagonal() ** 0.5
        # recondition to prevent NANs
        d[d < 1e-15] = 1
        # TODO, find optimal rcond!
        p = np.linalg.pinv(
            self.fisher_matrix_rel,
            rcond=self.rcond_cov,
        )
        cov = p / d.reshape(1, -1) / d.reshape(-1, 1)
        # uses only the length of resP, so resZ would be the same
        return (
            cov
            * self.residuals_average
            / (len(self.residuals_preferred) * 2 - cov.shape[0])
        )

    @depB_property
    def cor_matrix(self):
        cov = np.copy(self.cov_matrix)
        d = cov.diagonal() ** 0.5
        cov = cov / d.reshape(-1, 1)
        cov = cov / d.reshape(1, -1)
        return cov

    @depB_property
    def xfer_fit_error(self):
        cov = self.cov_matrix
        xfer, dfjac = self.transfer_jacobian_generate(xfer_fit_jacobian=True)
        dfjac = np.asarray(dfjac)
        perr_r = np.einsum(str("ij,ii,ij->j"), dfjac.real, cov, dfjac.real)
        perr_i = np.einsum(str("ij,ii,ij->j"), dfjac.imag, cov, dfjac.imag)

        dfjacAmp = np.einsum(str("ij,j->ij"), dfjac, abs(xfer) / xfer)
        perr_a = np.einsum(str("ij,ii,ij->j"), dfjacAmp.real, cov, dfjacAmp.real)
        dfjacRel = np.einsum(str("ij,j->ij"), dfjac, 1 / xfer)
        perr_ar = np.einsum(str("ij,ii,ij->j"), dfjacRel.real, cov, dfjacRel.real)
        perr_p = np.einsum(str("ij,ii,ij->j"), dfjacRel.imag, cov, dfjacRel.imag)
        cplx = perr_r ** 0.5 + 1j * perr_i ** 0.5
        return Bunch(
            cplx=cplx,
            mag=perr_a ** 0.5,
            mag_rel=perr_ar ** 0.5,
            mag_rel2=perr_a ** 0.5 / abs(xfer),
            phase=perr_p ** 0.5,
        )

    @depB_property
    def residuals_error(self):
        cov = self.cov_matrix
        res = self.residuals_preferred
        dfjac = self.residuals_jacobian
        dfjac = np.asarray(dfjac)
        perr_r = np.einsum(str("ij,ii,ij->j"), dfjac.real, cov, dfjac.real)
        perr_i = np.einsum(str("ij,ii,ij->j"), dfjac.imag, cov, dfjac.imag)

        dfjacAmp = np.einsum(str("ij,j->ij"), dfjac, abs(res) / res)
        perr_a = np.einsum(str("ij,ii,ij->j"), dfjacAmp.real, cov, dfjacAmp.real)
        dfjacRel = np.einsum(str("ij,j->ij"), dfjac, 1 / res)
        perr_ar = np.einsum(str("ij,ii,ij->j"), dfjacRel.real, cov, dfjacRel.real)
        perr_p = np.einsum(str("ij,ii,ij->j"), dfjacRel.imag, cov, dfjacRel.imag)
        cplx = perr_r ** 0.5 + 1j * perr_i ** 0.5
        return Bunch(
            cplx=cplx,
            mag=perr_a ** 0.5,
            mag_rel=perr_ar ** 0.5,
            phase=perr_p ** 0.5,
        )

    def transfer_generate(self, generate_inactive=False):
        if generate_inactive:
            # I believe these need a phasing adjustment if poles_overlay and zeros_overlay have different sizes
            # TODO also, if there are many poles_overlay and zeros_overlay, this is not numerically stable (should interlace)
            # TODO, should be using Xn_grid in the Z-domain
            xfer = self._extra_xfer_fit
            gain_effect = 1
            for coding_n, coding_d in itertools.zip_longest(
                self.coding_lists.num_ignore,
                self.coding_lists.den_ignore,
                fillvalue=None,
            ):
                if coding_n is not None:
                    xfer = xfer * coding_n.transfer()
                    gain_effect *= coding_n.gain_effect
                if coding_d is not None:
                    xfer = xfer / coding_d.transfer()
                    gain_effect /= coding_d.gain_effect
            return xfer * gain_effect
        else:
            xfer = self.xfer_inactive
            gain_effect = 1
            for coding_n, coding_d in itertools.zip_longest(
                self.coding_lists.num_active,
                self.coding_lists.den_active,
                fillvalue=None,
            ):
                if coding_n is not None:
                    xfer = xfer * coding_n.transfer()
                    gain_effect *= coding_n.gain_effect
                if coding_d is not None:
                    xfer = xfer / coding_d.transfer()
                    gain_effect /= coding_d.gain_effect
            return xfer * gain_effect

    def copy(self, **kwargs):
        return self.__class__(parent=self, **kwargs)

    def residuals_compute(self, data=None, xfer=None, W=None, rtype=None):
        if data is None:
            data = self.data
        if xfer is None:
            xfer = self.xfer_fit
        if W is None:
            W = self.W
        R = xfer / data
        return self.residuals_NLmap(R, W=W, rtype=rtype)

    def residuals_NLmap(self, R, W, rtype=None):
        if rtype is None:
            rtype = self.residuals_type
        if rtype == "zeros":
            return W * (R - 1)
        elif rtype == "poles":
            return W * (1 / R - 1)
        elif rtype == "dualA":
            return W * (R / 2 + 0.5 / R - 1)
        elif rtype == "dualB":
            return W * (R - 1 / R) / 2
        elif rtype == "log":
            R_abs_sq = TFmath.abs_sq(R)
            R_abs = R_abs_sq ** 0.5
            select = (R_abs <= 0) | ~np.isfinite(R_abs)

            if any(select):
                # TODO, make this report better, requires logging or some kind
                # of hook
                # import sys
                # print(
                #    "BADNESS AT: ",
                #    self.F_Hz[select],
                #    R_abs[select],
                #    np.sum(select) / len(select),
                #    file = sys.stderr
                # )
                # print(
                #    self.gain_coding.gain,
                # )
                pass

            with np.errstate(divide="ignore", invalid="ignore"):
                log_re = W * np.log(R_abs)
                # TODO, test if there is a way to scale the residuals_log using the real part
                # log_resq = np.sum(log_re**2)
                log_im = W * (R.imag / R_abs)
                if self.check_sign:
                    select = R.real < 0
                    log_im[select] = 2 - log_im[select]
            return log_re + 1j * log_im * self.residuals_log_im_scale
        return

    def transfer_jacobian_generate(
        self,
        xfer_fit_jacobian=False,
    ):
        # in principle it is faster to generate the xfer with the jacobian,
        # I am separating them as a wash
        # NEEDS SAME ORDERING AS PARAMETERS LIST

        # use separate bins for num and den so that the interlacing doesn't screw up the order
        op = Bunch(jac_list_num=[], jac_list_den=[], xfer_running=1)
        op.xfer_running = self.xfer_inactive
        rtype = self.residuals_type

        if xfer_fit_jacobian or rtype in ["poles", "zeros", "dualB", "dualA"]:

            def jac_modify(pjac, xfer, num):
                def generate(XY):
                    if num:
                        return XY * pjac / xfer
                    else:
                        return -XY * pjac / xfer

                return generate

        elif rtype == "log":

            def jac_modify(pjac, xfer, num):
                jac_abs_sq = TFmath.abs_sq(xfer)
                jac_re = (pjac.real * xfer.real + pjac.imag * xfer.imag) / jac_abs_sq

                def generate(XY):
                    Y = XY / xfer
                    XY_abs = abs(XY)
                    jac_im = (
                        self.residuals_log_im_scale
                        * ((pjac.imag * Y.real + pjac.real * Y.imag) - XY.imag * jac_re)
                        / XY_abs
                    )

                    if self.check_sign:
                        select = xfer.real < 0
                        jac_im[select] = -jac_im[select]

                    if num:
                        return jac_re + 1j * jac_im
                    else:
                        return -(jac_re + 1j * jac_im)

                return generate

        def coding_jac_append(coding, num):
            # don't include in jacobian if the codings_ignore contains the coding
            xfer, jac = coding.derivative_wtrans()

            if num:
                op.xfer_running = op.xfer_running * xfer * coding.gain_effect
                for pjac in jac:
                    op.jac_list_num.append(jac_modify(pjac, xfer, num=True))
            else:
                op.xfer_running = op.xfer_running / xfer / coding.gain_effect
                for pjac in jac:
                    op.jac_list_den.append(jac_modify(pjac, xfer, num=False))

        # interlace to prevent the value becoming crazy small and numerically unstable
        for coding_n, coding_d in itertools.zip_longest(
            self.coding_lists.num_active,
            self.coding_lists.den_active,
            fillvalue=None,
        ):
            if coding_n is not None:
                coding_jac_append(coding_n, num=True)
            if coding_d is not None:
                coding_jac_append(coding_d, num=False)

        xfer = op.xfer_running
        # have to do this after, otherwise they will be out of order from the interlacing
        jac_list = op.jac_list_num + op.jac_list_den
        # now need to actually inject the jacobian side of the chain rule, with the proper signs for the num, denom
        # first set are numerators of the xfer, including gain_coding
        if not xfer_fit_jacobian:
            if rtype == "poles":
                XY = xfer / self.data_effective
            elif rtype == "zeros":
                # minus sign to account for xfer on the denominator
                XY = -(self.data_effective / xfer)
            elif rtype == "dualA":
                # minus sign goes to plus on the second term
                XY = (xfer / self.data_effective - self.data_effective / xfer) / 2
            elif rtype == "dualB":
                # minus sign goes to plus on the second term
                XY = xfer / self.data_effective + self.data_effective / xfer
            elif rtype == "log":
                # minus sign goes to plus on the second term
                XY = xfer / self.data_effective
        else:
            XY = xfer

        for idx in range(len(jac_list)):
            jac_list[idx] = self.W * jac_list[idx](XY)

        return xfer, jac_list

    def optimize(self, **kwargs):
        # start with Nelder-Mead as it is fast, not needing the jacobian, then use Levenberg-Marquardt to seal the deal
        # self.optimize_NM(**kwargs)
        return self.optimize_LS(**kwargs)

    opt_xtol_default = 3e-16
    opt_ftol_default = 3e-5
    opt_gtol_default = 3e-5
    opt_method_default = None
    opt_rescale_default = True
    opt_nfev_default = 1000
    opt_log_default = False

    def optimize_LS(
        self,
        xtol=None,
        ftol=None,
        gtol=None,
        x_scale="jac",
        max_nfev=None,
        residuals_type=None,
        rescale=True,
        aid=None,
        log=None,
        **kwargs
    ):
        aid = ensure_aid(aid)

        args = [
            "optimize_LS_{arg}.MRF",
            "optimize_{arg}.MRF",
            "optimize_LS_{arg}",
            "optimize_{arg}",
        ]
        xtol = aid.hint_arg(xtol, args, default=self.opt_xtol_default, arg="xtol")
        ftol = aid.hint_arg(ftol, args, default=self.opt_ftol_default, arg="ftol")
        gtol = aid.hint_arg(gtol, args, default=self.opt_gtol_default, arg="gtol")
        max_nfev = aid.hint_arg(
            max_nfev, args, default=self.opt_nfev_default, arg="max_nfev"
        )

        rescale = rescale if rescale is not None else self.opt_rescale_default

        res_type_prev = self.residuals_type
        if residuals_type is not None:
            self.residuals_type = residuals_type
        if rescale:
            p_rescale = np.abs(np.asarray(self.parameters)) + 1e-5
        else:
            p_rescale = np.asarray(1)
        # TODO consider using auto_jacobian, gotta profile to see if it helps
        if np.any(abs(self.residuals_log_im_scale) > 1e-12):

            def opt(p):
                self.parameters = p * p_rescale
                return np.concatenate(
                    [
                        self.residuals_preferred.real,
                        self.residuals_preferred.imag,
                    ]
                )

            def opt_j(p, f=None):
                self.parameters = p * p_rescale
                jac = self.residuals_jacobian * p_rescale.reshape(-1, 1)
                return np.concatenate([jac.real, jac.imag], axis=1).T

        else:

            def opt(p):
                self.parameters = p * p_rescale
                return np.concatenate(
                    [
                        self.residuals_preferred.real,
                    ]
                )

            def opt_j(p, f=None):
                self.parameters = p * p_rescale
                jac = self.residuals_jacobian * p_rescale.reshape(-1, 1)
                return np.concatenate(
                    [
                        jac.real,
                    ],
                    axis=1,
                ).T

        # if using h_infinity, then wrap the residuals again and apply the ranked mask
        if self.h_infinity.shape != () or self.h_infinity > 0:
            prev_opt = opt
            prev_opt_j = opt_j

            def opt(p):
                resid = prev_opt(p)
                resid_r = resid[: len(self.F_Hz)]
                resid_i = resid[len(self.F_Hz) :]
                resid_sq = resid_r ** 2 + resid_i ** 2
                if self.h_infinity.shape == ():
                    if self.h_infinity_deweight == 0:
                        if self.h_infinity >= 1 - 1 / len(self.F_Hz):
                            idx_max = np.argmax(resid_sq)
                            select = np.array([idx_max])
                        else:
                            argsort = np.argsort(-resid_sq)
                            select = argsort[
                                : int((1 - self.h_infinity) * len(self.F_Hz) + 1)
                            ]
                        r = np.concatenate([resid_r[select], resid_i[select]])
                        return r
                    else:
                        argsort = np.argsort(-resid_sq)
                        select = argsort[
                            : int((1 - self.h_infinity) * len(self.F_Hz) + 1)
                        ]
                        resid_r[select] *= self.h_infinity_deweight
                        resid_i[select] *= self.h_infinity_deweight
                        return np.concatenate([resid_r, resid_i])
                else:
                    argsort = np.argsort(-resid_sq)
                    resid_r[argsort] *= self.h_infinity
                    resid_i[argsort] *= self.h_infinity
                    return np.concatenate([resid_r, resid_i])

            def opt_j(p, f=None):
                resid = prev_opt(p)
                resid_r = resid[: len(self.F_Hz)]
                resid_i = resid[len(self.F_Hz) :]
                resid_sq = resid_r ** 2 + resid_i ** 2
                j = prev_opt_j(p, f)
                if self.h_infinity.shape == ():
                    if self.h_infinity_deweight == 0:
                        if self.h_infinity >= 1 - 1 / len(self.F_Hz):
                            idx_max = np.argmax(resid_sq)
                            select = np.array([idx_max])
                        else:
                            argsort = np.argsort(-resid_sq)
                            select = argsort[
                                : int((1 - self.h_infinity) * len(self.F_Hz) + 1)
                            ]
                        select = np.concatenate([select, len(self.F_Hz) + select])
                        return j[select, :]
                    else:
                        argsort = np.argsort(-resid_sq)
                        select = argsort[
                            : int((1 - self.h_infinity) * len(self.F_Hz) + 1)
                        ]
                        j[select] *= self.h_infinity_deweight
                        j[len(self.F_Hz) + select] *= self.h_infinity_deweight
                        return j
                else:
                    argsort = np.argsort(-resid_sq)
                    j[argsort] *= self.h_infinity.reshape(-1, 1)
                    j[len(self.F_Hz) + argsort] *= self.h_infinity.reshape(-1, 1)
                    return j

        x0 = np.asarray(self.parameters) / p_rescale
        opt_r = scipy_optimize.trf(
            fun=opt,
            jac=opt_j,
            x0=x0,
            J0=opt_j(x0),
            f0=opt(x0),
            xtol=xtol,
            ftol=ftol,
            gtol=gtol,
            x_scale=x_scale,
            max_nfev=max_nfev,
            verbose=0,
            **kwargs
        )

        # don't actually return an OptimizeResult because it kills matlab
        results = Bunch()
        results.update(opt_r)
        self.parameters = results.x * p_rescale
        log = aid.hint_arg(log, args, default=self.opt_log_default, arg="log")
        if log:
            aid.log(results.message, results.nfev, results.njev)
        if residuals_type is not None:
            self.residuals_type = res_type_prev
        return results

    def optimize_NM(
        self,
        bootstrap=True,
        tol=3e-16,
        xtol=None,
        ftol=None,
        method="Nelder-Mead",
        max_nfev=None,
        residuals_type=None,
        rescale=True,
        aid=None,
        **kwargs
    ):
        """
        Uses Nelder-Mead with a simplex that is bootstrapped from the
        svd of the jacobian. This gives particularly good
        initial directions to check
        """
        aid = ensure_aid(aid)
        if bootstrap:
            jac = np.concatenate(
                [self.residuals_jacobian.real, self.residuals_jacobian.imag], axis=1
            )
            res = np.concatenate(
                [self.residuals_preferred.real, self.residuals_preferred.imag], axis=0
            )
            d = np.sum(jac * jac, axis=1) ** 0.5
            # p_rescale the jacobian using the norms, makes the sum of the singular values equal to the number
            # of parameters (and greatly improves conditioning)
            d[abs(d) < 1e-10] = 1
            jac /= d.reshape(-1, 1)
            u, s, v = np.linalg.svd(jac, full_matrices=False)
            n_r = (res.dot(res)) ** 0.5
            v_r = v.dot(res) / n_r
            # deweight the smallest singular value as it will be replaced and it shouldn't screw up the fit if it is nasty
            # TODO, handle bad singular values
            u_proj = u * (v_r * n_r / (s + 1e-4 + np.min(s)) / d).reshape(-1, 1)
            u_best = np.sum(u_proj, axis=1)
            # now replace the smallest singular vector with the optimal nearest quadratic solution
            u_proj[:, -1] = u_best
            initial_simplex = np.concatenate(
                [np.zeros_like(d).reshape(-1, 1), u_proj], axis=1
            ).T + np.asarray(self.parameters)
        else:
            initial_simplex = None

        res_type_prev = self.residuals_type
        if residuals_type is not None:
            self.residuals_type = residuals_type
        try:

            args = [
                "optimize_NM_{arg}.MRF",
                "optimize_{arg}.MRF",
                "optimize_NM_{arg}",
                "optimize_{arg}",
            ]
            xtol = aid.hint_arg(xtol, args, default=self.opt_xtol_default, arg="xtol")
            ftol = aid.hint_arg(ftol, args, default=self.opt_ftol_default, arg="ftol")
            max_nfev = aid.hint_arg(
                max_nfev, args, default=self.opt_nfev_default, arg="max_nfev"
            )
            rescale = rescale if rescale is not None else self.opt_rescale_default

            if rescale:
                p_rescale = np.abs(np.asarray(self.parameters)) + aid.hint(
                    "MRF_rescale_min", default=1e-5
                )
            else:
                p_rescale = np.asarray(1)

            if residuals_type is not None:
                self.residuals_type = residuals_type

            def opt(p):
                self.parameters = p * p_rescale
                return TFmath.norm_sq(self.residuals_preferred)

            if self.h_infinity.shape != () or self.h_infinity > 0:
                prev_opt = opt

                def opt(p):
                    resid = prev_opt(p)
                    resid_r = resid[: len(self.F_Hz)]
                    resid_i = resid[len(self.F_Hz) :]
                    resid_sq = resid_r ** 2 + resid_i ** 2
                    if self.h_infinity.shape == ():
                        if self.h_infinity_deweight == 0:
                            if self.h_infinity >= 1 - 1 / len(self.F_Hz):
                                idx_max = np.argmax(resid_sq)
                                select = np.array([idx_max])
                            else:
                                argsort = np.argsort(-resid_sq)
                                select = argsort[
                                    : int((1 - self.h_infinity) * len(self.F_Hz) + 1)
                                ]
                            r = np.concatenate([resid_r[select], resid_i[select]])
                            return r
                        else:
                            argsort = np.argsort(-resid_sq)
                            select = argsort[
                                : int((1 - self.h_infinity) * len(self.F_Hz) + 1)
                            ]
                            resid_r[select] *= self.h_infinity_deweight
                            resid_i[select] *= self.h_infinity_deweight
                            return np.concatenate([resid_r, resid_i])
                    else:
                        argsort = np.argsort(-resid_sq)
                        resid_r[argsort] *= self.h_infinity
                        resid_i[argsort] *= self.h_infinity
                        return np.concatenate([resid_r, resid_i])

            kwargs["maxiter"] = max_nfev
            opt_r = scipy_optimize.neldermead(
                func=opt,
                x0=np.asarray(self.parameters) / p_rescale,
                xatol=xtol,
                fatol=ftol,
                maxiter=max_nfev,
                maxfev=max_nfev,
                initial_simplex=initial_simplex,
                callback=None,
                disp=False,
                return_all=False,
            )
            results = Bunch()
            results.update(opt_r)
            self.parameters = results.x * p_rescale

            if residuals_type is not None:
                self.residuals_type = res_type_prev

            return results
        finally:
            if residuals_type is not None:
                self.residuals_type = res_type_prev

    def optimize_min(
        self,
        tol=3e-16,
        xtol=None,
        ftol=None,
        method="Nelder-Mead",
        max_nfev=None,
        residuals_type=None,
        rescale=True,
        aid=None,
        **kwargs
    ):
        res_type_prev = self.residuals_type
        aid = ensure_aid(aid)

        args = [
            "optimize_min_{arg}.MRF",
            "optimize_{arg}.MRF",
            "optimize_min_{arg}",
            "optimize_{arg}",
        ]
        xtol = aid.hint_arg(xtol, args, default=self.opt_xtol_default, arg="xtol")
        ftol = aid.hint_arg(ftol, args, default=self.opt_ftol_default, arg="ftol")
        max_nfev = aid.hint_arg(
            max_nfev, args, default=self.opt_nfev_default, arg="max_nfev"
        )

        method = method if method is not None else self.opt_method_default
        rescale = rescale if rescale is not None else self.opt_rescale_default

        if rescale:
            p_rescale = np.abs(np.asarray(self.parameters)) + aid.hint(
                "MRF_rescale_min", default=1e-5
            )
        else:
            p_rescale = np.asarray(1)

        if residuals_type is not None:
            self.residuals_type = residuals_type

        def opt(p):
            self.parameters = p * p_rescale
            return TFmath.norm_sq(self.residuals_preferred)

        def opt_j(p):
            self.parameters = p
            jac = self.residuals_jacobian * p_rescale.reshape(-1, 1)
            return 2 * (
                jac.real.dot(self.residuals_preferred.real)
                + jac.imag.dot(self.residuals_preferred.imag)
            )

        def opt_h(p):
            self.parameters = p
            jac = self.residuals_jacobian * p_rescale.reshape(-1, 1)
            hess = 2 * (jac.real.dot(jac.T.real) + jac.imag.dot(jac.T.imag))
            return hess

        more_args = dict()
        if method in ["Nelder-Mead"]:
            # doesn't need jac or hess
            pass
        else:
            more_args["jac"] = opt_j
            more_args["hess"] = opt_h

        kwargs["maxiter"] = max_nfev
        import scipy.optimize

        opt_r = scipy.optimize.minimize(
            opt,
            np.asarray(self.parameters) / p_rescale,
            tol=tol,
            method=method,
            options=kwargs,
            **more_args
        )
        results = Bunch()
        results.update(opt_r)
        self.parameters = results.x * p_rescale

        if residuals_type is not None:
            self.residuals_type = res_type_prev

        return results

    def xfer_eval(self, F_Hz):
        return self.ZPKrep.xfer_eval(
            F_Hz=F_Hz,
        )

    def distance_limit(self, F_Hz, with_derivative=False):
        F_idx = np.searchsorted(self.F_Hz, F_Hz) - 1
        if F_idx < 1:
            val = self.F_Hz[1] - self.F_Hz[0]
            assert val > 0
            if not with_derivative:
                return val
            else:
                return val, 0
        if F_idx >= len(self.F_Hz) - 2:
            val = F_Hz - self.F_Hz[-4]
            # TODO
            # val = self.F_Hz[-1] - self.F_Hz[-4]
            assert val > 0
            if not with_derivative:
                return val
            else:
                return val, 1
        # get the local mean width of bins to restrict bandwidths
        F_mean_Hz_2 = (self.F_Hz[F_idx + 2] - self.F_Hz[F_idx]) / 2
        F_mean_Hz_1 = (self.F_Hz[F_idx + 1] - self.F_Hz[F_idx - 1]) / 2
        x = (F_Hz - self.F_Hz[F_idx]) / (self.F_Hz[F_idx + 1] - self.F_Hz[F_idx])
        if self.data is None:
            ratio = 1
        else:
            W = self.W
            if W is None:
                W = 1
            if np.asarray(W).shape != self.data.shape:
                W = 5
            else:
                W = W[F_idx - 1 : F_idx + 3]
            dabs = abs(self.data[F_idx - 1 : F_idx + 3])
            y_max = max(abs((1 - 1 / W)) * dabs)
            y_min = min(abs((1 + 1 / W)) * dabs)
            ratio = y_min / y_max
            if ratio > 1:
                ratio = 1
        # TODO
        # use the SNR to make a local density estimate!
        val = F_mean_Hz_1 * (1 - x) + x * F_mean_Hz_2
        if val < 0:
            print("VAL: ", F_mean_Hz_1, F_mean_Hz_2, x)
            assert False
        assert self.distance_limit_scale > 0
        assert ratio > 0

        if not with_derivative:
            return ratio * val * self.distance_limit_scale
        else:
            d = (F_mean_Hz_2 - F_mean_Hz_1) / (self.F_Hz[F_idx + 1] - self.F_Hz[F_idx])
            return (
                ratio * val * self.distance_limit_scale,
                ratio * d * self.distance_limit_scale,
            )

    def optimize_delay(self, only_delay=True, **kwargs):
        if self.delay_coding is None:
            return
        # store the previous so they may be restored
        prev_codings_ignore = self.codings_ignore

        if only_delay:
            # ignore all other codings
            self.codings_ignore = set(self.num_codings + self.den_codings)

        ret = self.optimize(**kwargs)

        # restore
        self.codings_ignore = prev_codings_ignore
        return ret

    def codings_in_region(
        self,
        F_min_Hz,
        F_max_Hz,
        num=True,
        den=True,
    ):
        raise NotImplementedError()
        # TODO, make compatible with S domain
        # TODO: add more tests and bandwidth limits
        gotten = []

        def test_get(coding):
            r_lst = coding.roots_c_Sf(self)
            if len(r_lst) == 1:
                r = r_lst[0]
                F = np.angle(r) / np.pi * self.F_nyquist_Hz
                if F_max_Hz is not None and F > F_max_Hz:
                    return
                if F_min_Hz is not None and F < F_min_Hz:
                    return
                gotten.append(coding)

        if num:
            for coding in self.num_codings:
                test_get(coding)
        if den:
            for coding in self.den_codings:
                test_get(coding)
        return gotten

    def regenerate(
        self,
        ZPKrep=None,
        zeros=None,
        poles=None,
        gain=None,
        p_c=None,
        z_c=None,
        p_r=None,
        z_r=None,
        coding_map=None,
        **kwargs
    ):
        from .ZPKrep2MRF import MRF2MRF

        return MRF2MRF(
            self,
            ZPKrep=ZPKrep,
            zeros=zeros,
            poles=poles,
            gain=gain,
            p_c=p_c,
            z_c=z_c,
            p_r=p_r,
            z_r=z_r,
            coding_map=coding_map,
            **kwargs
        )

    @contextlib.contextmanager
    def with_codings_only(self, codings, prev_ignores=True):
        prev_ignores = self.codings_ignore
        if prev_ignores:
            self.codings_ignore = self.codings_ignore.union(
                self.root_codings_set - set(codings)
            )
        else:
            self.codings_ignore = self.root_codings_set - set(codings)
        yield
        self.codings_ignore = prev_ignores
        return

    @contextlib.contextmanager
    def without_codings_only(self, codings, prev_ignores=True):
        prev_ignores = self.codings_ignore
        if prev_ignores:
            self.codings_ignore = self.codings_ignore.union(set(codings))
        else:
            self.codings_ignore = set(codings)
        yield
        self.codings_ignore = prev_ignores


def check_jac(fitterMRF, idx, scaling=0.01):
    p_orig = np.asarray(fitterMRF.parameters)
    jac = fitterMRF.residuals_jacobian
    res1 = fitterMRF.residuals_preferred
    shift = [0] * len(p_orig)
    offset = scaling / (np.dot(jac[idx], jac[idx].conjugate()).real) ** 0.5
    shift[idx] = offset
    fitterMRF.parameters = p_orig + np.asarray(shift)
    res2 = fitterMRF.residuals_preferred
    njac = (res2 - res1) / offset

    fitterMRF.parameters = p_orig
    with np.errstate(divide="ignore", invalid="ignore"):
        return njac / jac[idx], TFmath.abs_sq(jac[idx]) > 0


class MultiReprFilterZ(MultiReprFilterBase, DataFiltFitZ):
    pass


class MultiReprFilterS(MultiReprFilterBase, DataFiltFitSf):
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
Utilities to manipulate ZPK roots S to/from Z, and make transfer functions
"""

import numpy as np
from wield.bunch import Bunch


def analytic_translation(
    F_Hz,
    time_s=None,
    F_displacement_Hz=0,
):
    """
    time_bins shou
    """
    if time_s is None:
        time_s = (
            np.linspace(
                0, (len(F_Hz) - 1) / np.max(F_Hz) / 2, 3 * (1 + 1 * len(F_Hz) / 2)
            )
            * 1.2
        )
    if isinstance(time_s, (float, int)):
        time_s = (
            np.linspace(
                0, (len(F_Hz) - 1) / np.max(F_Hz) / 2, time_s * (1 + 1 * len(F_Hz) / 2)
            )
            * 1.2
        )

    max_F_Hz = np.max(F_Hz)

    t, f = np.meshgrid(time_s, F_Hz)
    fourier_c = np.cos(2 * np.pi * f * t)
    fourier_s = np.sin(2 * np.pi * f * t)

    disp_factor = np.exp(-F_displacement_Hz * time_s).reshape(-1, 1)
    # u_c, s_c, v_c = np.linalg.svd(fourier_c)
    # u_s, s_s, v_s = np.linalg.svd(fourier_s)
    fourier_c_inv = disp_factor * np.linalg.pinv(fourier_c)
    fourier_s_inv = disp_factor * np.linalg.pinv(fourier_s)

    trans_c = np.dot(fourier_c, fourier_c_inv)
    trans_s = np.dot(fourier_s, fourier_s_inv)

    trans_sc = np.dot(fourier_s, fourier_c_inv)
    trans_cs = np.dot(fourier_c, fourier_s_inv)

    def apply_translation(cplx_vect):
        return np.dot(trans_c, cplx_vect.real) + 1j * np.dot(trans_s, cplx_vect.imag)

    def apply_continuation(cplx_vect):
        return -(
            1j * np.dot(trans_sc, cplx_vect.real) + np.dot(trans_cs, cplx_vect.imag)
        )

    def stability_test(cplx_vect):
        return apply_translation(cplx_vect) / apply_continuation(cplx_vect)

    return Bunch(locals())


def analytic_translation_dual(
    F_Hz,
    time_s=None,
    F_displacement_Hz=0,
):
    """
    time_bins shou
    """
    if time_s is None:
        time_s = (
            np.linspace(
                0, (len(F_Hz) - 1) / np.max(F_Hz) / 2, 3 * (1 + 1 * len(F_Hz) / 2)
            )
            * 1.2
        )
    if isinstance(time_s, (float, int)):
        time_s = (
            np.linspace(
                0, (len(F_Hz) - 1) / np.max(F_Hz) / 2, time_s * (1 + 1 * len(F_Hz) / 2)
            )
            * 1.2
        )

    max_F_Hz = np.max(F_Hz)

    t, f = np.meshgrid(time_s, F_Hz)
    fourier_c = np.cos(2 * np.pi * f * t)
    fourier_s = np.sin(2 * np.pi * f * t)
    fourier_cs = np.hstack([fourier_c, fourier_s])
    fourier_sc = np.hstack([fourier_s, -fourier_c])
    print(fourier_cs.shape)

    disp_factor = np.exp(-F_displacement_Hz * time_s).reshape(-1, 1)
    print(disp_factor.shape)
    disp_factor = np.vstack([disp_factor, disp_factor])
    print(disp_factor.shape)

    fourier_cs_inv = disp_factor * np.linalg.pinv(fourier_cs)

    trans = np.dot(fourier_cs, fourier_cs_inv)
    trans_sc = np.dot(fourier_sc, fourier_cs_inv)

    def apply_translation(cplx_vect):
        return np.dot(trans, cplx_vect)

    def apply_continuation(cplx_vect):
        return -1j * np.dot(trans_sc, cplx_vect)

    return Bunch(locals())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
Contains setup functions for test data, these are returned in a QuickBunch dictionary, with some annotation about the number of data sets
"""
import numpy as np
from wield.bunch import Bunch

from ..utilities.np import logspaced
from . import utilities


def meta_generator(
    SNR=10,
    log=False,
    N=1000,
    Nreal_max=10,
    Ncplx_max=10,
):
    def dataset_gen(sN=0, iN=0, **kwargs):
        rand = np.random.RandomState()
        rand.seed(sN)

        Nreal = rand.randint(Nreal_max) + 1
        Ncplx = rand.randint(Ncplx_max) + 1

        NrealP = (Nreal + 1) // 2  # rand.randint(Nreal//2 + 1) + (Nreal + 1)//4
        NrealZ = Nreal - NrealP
        NcplxP = (Ncplx + 1) // 2  # (Ncplx + 1)//4 + rand.randint((Ncplx + 1)//2 + 1)
        NcplxZ = Ncplx - NcplxP

        F_max = 100
        F_max_R = 80
        F_min = 0.001
        F_nyquist_Hz = None
        if log:
            F_Hz = logspaced(F_min, F_max, N)

            def densityBW(F_Hz):
                return np.log(F_max / F_min) * F_Hz / N

            Preal = -F_min * np.exp(
                rand.uniform(0, 1, NrealP) * np.log(F_max_R / F_min)
            )
            PcplxF = (
                2
                * F_min
                * np.exp(rand.uniform(0, 1, NcplxP) * np.log(F_max_R / F_min / 2))
            )
            PcplxR = -np.exp(rand.uniform(0, 1, NcplxP)) * densityBW(PcplxF)
            Pcplx = PcplxR + 1j * PcplxF

            Zreal = -F_min * np.exp(
                rand.uniform(0, 1, NrealZ) * np.log(F_max_R / F_min)
            )
            ZcplxF = (
                2
                * F_min
                * np.exp(rand.uniform(0, 1, NcplxZ) * np.log(F_max_R / F_min / 2))
            )
            ZcplxR = -np.exp(rand.uniform(0, 2, NcplxZ)) * densityBW(ZcplxF)
            Zcplx = ZcplxR + 1j * ZcplxF
        else:
            F_Hz = np.linspace(0, F_max, N)

            def densityBW(F_Hz):
                return (F_max - F_min) / N

            Preal = -F_min * np.exp(
                rand.uniform(0, 1, NrealP) * np.log(F_max_R / F_min)
            )
            Zreal = -F_min * np.exp(
                rand.uniform(0, 1, NrealZ) * np.log(F_max_R / F_min)
            )
            PcplxF = rand.uniform(0, F_max_R, NcplxP)
            ZcplxF = rand.uniform(0, F_max_R, NcplxZ)
            PcplxR = -rand.uniform(0, densityBW(PcplxF), NcplxP)
            ZcplxR = -rand.uniform(0, densityBW(ZcplxF), NcplxZ)
            Pcplx = PcplxR + 1j * PcplxF
            Zcplx = ZcplxR + 1j * ZcplxF
        K = 1
        Z = np.concatenate([Zreal, Zcplx, Zcplx.conjugate()])
        P = np.concatenate([Preal, Pcplx, Pcplx.conjugate()])

        Rp = np.prod(abs(Pcplx)) * np.prod(abs(Preal))
        Rz = np.prod(abs(Zcplx)) / np.prod(abs(Zreal))
        R = Rp / Rz
        print("RATIO1", R, Rp, Rz)
        if abs(np.log10(Rp)) > 5:
            Pcplx /= Rp ** (1.0 / len(Pcplx))
        if abs(np.log10(Rz)) > 5:
            Zcplx /= Rz ** (1.0 / len(Zcplx))
        Rp = np.prod(abs(Pcplx)) * np.prod(abs(Preal))
        Rz = np.prod(abs(Zcplx)) / np.prod(abs(Zreal))
        R = Rp / Rz
        print("RATIO2", R, Rp, Rz)

        return utilities.generator_autofill(
            F_Hz=F_Hz,
            SNR=SNR,
            F_nyquist_Hz=F_nyquist_Hz,
            bestfit_ZPK_s=[Z, P, K],
            sN=sN,
            iN=iN,
        )

    return dataset_gen


datasets = Bunch()

datasets.rand10_log1k = utilities.make_description(
    generator=meta_generator(
        SNR=10,
        log=True,
        N=1000,
        Nreal_max=5,
        Ncplx_max=10,
    ),
    sets=-1,
    instances=-1,
)


datasets.rand10_lin1k = utilities.make_description(
    generator=meta_generator(
        SNR=10,
        log=False,
        N=1000,
        Nreal_max=5,
        Ncplx_max=10,
    ),
    sets=-1,
    instances=-1,
)


datasets.rand6_log100 = utilities.make_description(
    generator=meta_generator(
        SNR=10,
        log=True,
        N=100,
        Nreal_max=4,
        Ncplx_max=6,
    ),
    sets=-1,
    instances=-1,
)


datasets.rand6_lin100 = utilities.make_description(
    generator=meta_generator(
        SNR=10,
        log=False,
        N=100,
        Nreal_max=4,
        Ncplx_max=6,
    ),
    sets=-1,
    instances=-1,
)


datasets.rand2_log100 = utilities.make_description(
    generator=meta_generator(
        SNR=10,
        log=True,
        N=100,
        Nreal_max=2,
        Ncplx_max=2,
    ),
    sets=-1,
    instances=-1,
)


datasets.rand2_lin100 = utilities.make_description(
    generator=meta_generator(
        SNR=10,
        log=False,
        N=100,
        Nreal_max=2,
        Ncplx_max=2,
    ),
    sets=-1,
    instances=-1,
)

datasets.rand6_log100E = utilities.make_description(
    generator=meta_generator(
        SNR=None,
        log=True,
        N=100,
        Nreal_max=4,
        Ncplx_max=6,
    ),
    sets=-1,
    instances=-1,
)


datasets.rand6_lin100E = utilities.make_description(
    generator=meta_generator(
        SNR=None,
        log=False,
        N=100,
        Nreal_max=6,
        Ncplx_max=6,
    ),
    sets=-1,
    instances=-1,
)

datasets.rand2_log100E = utilities.make_description(
    generator=meta_generator(
        SNR=None,
        log=True,
        N=100,
        Nreal_max=2,
        Ncplx_max=2,
    ),
    sets=-1,
    instances=-1,
)


datasets.rand2_lin100E = utilities.make_description(
    generator=meta_generator(
        SNR=None,
        log=False,
        N=100,
        Nreal_max=2,
        Ncplx_max=2,
    ),
    sets=-1,
    instances=-1,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


import numpy as np
from matplotlib import gridspec

from ..utilities.np import logspaced
from ..utilities.mpl import generate_stacked_plot_ax
from ..utilities import args

from .utilities import ZPKrep2Sf
from . import plot_augment


def plot_fit(
    self,
    fitter,
    plot_zp=True,
    plot_past_data=False,
    plot_to_nyquist=True,
    plot_data_scale=10,
    plot_data_scale_low=2,
    with_error=True,
    with_data_error=True,
    plot_data=True,
    N_points=150,
    gs_base=gridspec.GridSpec(1, 1)[0],
    fig=None,
    ax_bunch=None,
    xscale=None,
    plot_now=True,
    label_data=args.UNSPEC,
    label_fit=args.UNSPEC,
    residuals=False,
):
    use_phase = True
    label_data = args.argscan(label_data, self.label_data, "data")
    label_fit = args.argscan(
        label_fit, self.label_fit, "fit ({nP}P, {nZ}Z, Rsq ={Rsq:.2e})"
    )
    # if fitter.magsq_data:
    #    use_phase = False
    if fitter.phase_missing:
        use_phase = False

    if ax_bunch is None:
        axB = generate_stacked_plot_ax(
            [
                ("mag", True),
                ("phase", use_phase),
            ],
            height_ratios=dict(
                mag=2,
            ),
            heights_phys_in_default=1,
            xscales="log",
            gs_base=gs_base,
            fig=fig,
            ax_bunch=ax_bunch,
        )
    else:
        axB = ax_bunch

    ZPKrepSf = ZPKrep2Sf(fitter.ZPKrep)
    if len(ZPKrepSf.poles) + len(ZPKrepSf.zeros) == 0:
        plot_zp = False
    if plot_zp:
        axB.mag2 = axB.mag.twinx()

    data_list = []
    if plot_data and fitter.data is not None:
        axB.plot_data = plot_augment.plot_augment_data(
            self,
            ax_mag=axB.mag,
            ax_phase=axB.phase,
            data_list=data_list,
            with_error=with_error,
            label=label_data,
            residuals=residuals,
        )
        if plot_now:
            axB.plot_data(fitter)

    fit_list = []
    if plot_past_data:
        intersperse = min((len(fitter.F_Hz) - 1) // N_points + 1, 1)
        if fitter.F_nyquist_Hz is not None:
            if plot_to_nyquist:
                f = np.sort(
                    np.concatenate(
                        [
                            np.linspace(0, fitter.F_nyquist_Hz, N_points // 2),
                            logspaced(
                                np.min(fitter.F_Hz[fitter.F_Hz > 0])
                                / plot_data_scale_low,
                                fitter.F_nyquist_Hz,
                                N_points // 2,
                            ),
                            fitter.F_Hz[intersperse // 2 :: intersperse],
                        ]
                    )
                )
            else:
                F_max_Hz = np.max(fitter.F_Hz) * plot_data_scale
                f = np.sort(
                    np.concatenate(
                        [
                            np.linspace(0, F_max_Hz, N_points // 2),
                            logspaced(
                                np.min(fitter.F_Hz[fitter.F_Hz > 0])
                                / plot_data_scale_low,
                                F_max_Hz,
                                N_points // 2,
                            ),
                            fitter.F_Hz[intersperse // 2 :: intersperse],
                        ]
                    )
                )
        else:
            F_max_Hz = np.max(fitter.F_Hz) * plot_data_scale
            f = np.sort(
                np.concatenate(
                    [
                        np.linspace(0, F_max_Hz, N_points // 2),
                        logspaced(
                            np.min(fitter.F_Hz[fitter.F_Hz > 0]) / plot_data_scale_low,
                            F_max_Hz,
                            N_points // 2,
                        ),
                        fitter.F_Hz[intersperse // 2 :: intersperse],
                    ]
                )
            )
        axB.plot_fit = plot_augment.plot_augment_fit(
            self,
            ax_mag=axB.mag,
            ax_phase=axB.phase,
            with_error=with_error,
            label=label_fit,
            F_Hz=f,
            fit_list=fit_list,
            residuals=residuals,
        )
    else:
        axB.plot_fit = plot_augment.plot_augment_fit(
            self,
            ax_mag=axB.mag,
            ax_phase=axB.phase,
            with_error=with_error,
            label=label_fit,
            fit_list=fit_list,
            residuals=residuals,
        )
    if plot_now:
        axB.plot_fit(fitter)

    if plot_zp:
        axB.plot_zp = plot_augment.plot_augment_zp(self, ax=axB.mag2)
        if plot_now:
            axB.plot_zp(fitter)
    else:
        axB.plot_zp = lambda fitter: None

    axB.mag.set_ylabel("Magnitude")
    if axB.phase:
        axB.phase.set_ylabel("Phase [deg]")
    axB.ax_bottom.set_xlabel("Frequency [Hz]")

    def finalize():
        min_lims = []
        max_lims = []
        for fitB in data_list:
            min_err_lim = min(
                fitB.min_data_err_lim,
                max(fitB.min_data_err_lim / 30, fitB.min_data_err_lim * 1),
            )
            min_lims.append(min_err_lim)
            max_err_lim = max(
                fitB.max_data_err_lim,
                min(fitB.max_data_err_lim * 30, fitB.max_data_err_lim / 1),
            )
            max_lims.append(max_err_lim)
        for fitB in fit_list:
            min_err_lim = min(
                fitB.median_fit_lim,
                fitB.min_fit_err_lim,
                max(fitB.min_fit_err_lim / 30, fitB.min_fit_err_lim * 1),
            )
            min_lims.append(min_err_lim)
            max_err_lim = max(
                fitB.median_fit_lim,
                fitB.max_fit_err_lim,
                min(fitB.max_fit_err_lim * 30, fitB.max_fit_err_lim / 1),
            )
            max_lims.append(max_err_lim)
        min_lims = np.asarray(min_lims)
        max_lims = np.asarray(max_lims)
        axB.mag.set_ylim(
            np.min(min_lims[np.isfinite(min_lims)]),
            np.max(max_lims[np.isfinite(max_lims)]),
        )

        scales = []
        mins = []
        minsNZ = []
        maxs = []
        for fitB in fit_list + data_list:
            scales.append(fitB.scale_log)
            mins.append(fitB.minF_Hz)
            minsNZ.append(fitB.minF_HzNZ)
            maxs.append(fitB.maxF_Hz)

        if xscale is None:
            if np.count_nonzero(scales) / len(scales) > 0.5:
                final_xscale = "log_zoom"
            else:
                final_xscale = "linear"
        else:
            final_xscale = xscale

        if final_xscale in ["log", "log_zoom"]:
            xmin = np.min(minsNZ)
        else:
            xmin = np.min(mins)
        xmax = np.max(maxs)
        axB.ax_bottom.set_xscale(final_xscale)
        axB.ax_bottom.set_xlim(xmin, xmax)
        for subax in axB.ax_list_0:
            subax.set_xscale(final_xscale)
            subax.set_xlim(xmin, xmax)

        axB.mag.legend(ncol=1, fontsize=8, loc="best")

    axB.finalizers.append(finalize)
    return axB



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


import scipy
import scipy.linalg

try:
    # if the package was built then we can go faster
    from .lapack.svd import svd as svd_fast
except ImportError:
    svd_fast = None


def SVD_SV(
    a,
    overwrite_a=False,
    n_smallest=None,
):
    if svd_fast is not None:
        if n_smallest is not None:
            S, V = svd_fast(
                a,
                compute_u=False,
                compute_v=True,
                il=-n_smallest,
                overwrite_a=overwrite_a,
            )
        else:
            S, V = svd_fast(
                a,
                compute_u=False,
                compute_v=True,
                overwrite_a=overwrite_a,
            )
    else:
        U, S, V = scipy.linalg.svd(
            a,
            compute_uv=True,
            full_matrices=False,
            overwrite_a=overwrite_a,
        )
    return S, V



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""

import numpy as np
import scipy
import scipy.signal
from . import order_reduce


def ss2zpk(
    A,
    B,
    C,
    D,
    E=None,
    idx_in=None,
    idx_out=None,
    Q_rank_cutoff=1e-5,
    Q_rank_cutoff_unstable=None,
    F_match_Hz=1e-10,
    fmt="IIRrational",
):
    if idx_in is None:
        if B.shape[1] == 1:
            idx_in = 0
        else:
            raise RuntimeError("Must specify idx_in if B indicates MISO/MIMO system")
    if idx_out is None:
        if C.shape[0] == 1:
            idx_out = 0
        else:
            raise RuntimeError("Must specify idx_in if C indicates SIMO/MIMO system")
    B = B[:, idx_in : idx_in + 1]
    C = C[idx_out : idx_out + 1, :]
    D = D[idx_out : idx_out + 1, idx_in : idx_in + 1]

    if E is None:
        p = scipy.linalg.eig(A, left=False, right=False)
    else:
        p = scipy.linalg.eig(A, E, left=False, right=False)
    SS = np.block([[A, B], [C, D]])
    if E is None:
        z = scipy.linalg.eig(
            a=SS,
            b=np.diag(np.concatenate([np.ones(A.shape[0]), np.zeros(1)])),
            left=False,
            right=False,
        )
    else:
        SSE = np.block(
            [
                [E, np.zeros(E.shape[0]).reshape(-1, 1)],
                [np.zeros(E.shape[1]).reshape(1, -1), np.zeros(1).reshape(1, 1)],
            ]
        )
        z = scipy.linalg.eig(a=SS, b=SSE, left=False, right=False)
    z = np.asarray([_ for _ in z if np.isfinite(_.real)])
    k = 1
    z, p, k = order_reduce.order_reduce_zpk(
        (z, p, k),
        reduce_c=True,
        reduce_r=True,
        Q_rank_cutoff=Q_rank_cutoff,
        Q_rank_cutoff_unstable=Q_rank_cutoff_unstable,
    )
    s_match_wHz = F_match_Hz * 2j * np.pi
    tf0 = (
        np.matmul(C, np.matmul(np.linalg.inv(np.eye(A.shape[0]) * s_match_wHz - A), B))
        + D
    )[..., 0, 0]
    w, zpk0 = scipy.signal.freqs_zpk(z, p, k, s_match_wHz)
    k = abs(tf0 / zpk0)

    if fmt == "IIRrational":
        z = np.asarray(z) / (2 * np.pi)
        p = np.asarray(p) / (2 * np.pi)
        k = np.asarray(k) * (2 * np.pi) ** (len(z) - len(p))
    elif fmt == "scipy":
        pass
    else:
        raise RuntimeError("Unrecognized fmt parameter")
    return z, p, k


def ss2xfer(A, B, C, D, E=None, F_Hz=None, idx_in=None, idx_out=None):
    if idx_in is None:
        if B.shape[1] == 1:
            idx_in = 0
        else:
            raise RuntimeError("Must specify idx_in if B indicates MISO/MIMO system")
    if idx_out is None:
        if C.shape[0] == 1:
            idx_out = 0
        else:
            raise RuntimeError("Must specify idx_in if C indicates SIMO/MIMO system")
    B = B[:, idx_in : idx_in + 1]
    C = C[idx_out : idx_out + 1, :]
    D = D[idx_out : idx_out + 1, idx_in : idx_in + 1]

    s = 2j * np.pi * F_Hz
    if E is None:
        S = np.eye(A.shape[0]).reshape(1, *A.shape) * s.reshape(-1, 1, 1)
        return (
            np.matmul(C, np.matmul(np.linalg.inv(S - A.reshape(1, *A.shape)), B)) + D
        )[..., 0, 0]
    else:
        return (
            np.matmul(
                C,
                np.matmul(
                    np.linalg.inv(E * s.reshape(-1, 1, 1) - A.reshape(1, *A.shape)), B
                ),
            )
            + D
        )[..., 0, 0]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
These are chebychev polynomials on a rotated domain. While they should be the chebychev polynomials form
F(ix) = Sum_i c_i * C_i(x / i)

so that evaluated along the imaginary line they are standard chebychev's, these are not those. Instead they are the chebychevs of

F(ix) = Sum_i c_i * (C_ei(x) + iC_oi(x))

such that all of the c_i coefficients are pure real, but the even polynomials provide the real part and the odd ones the imaginary part. This is easy to create a vandermonde matrix from and easy to solve. The roots of this form of polynomial will be paired left/right over the imaginary line rather than paired with conjugates as for pure real coefficients.

To relate the coefficients to the original requires them to be rotated 90 degrees
"""


import numpy as np
from . import chebychev

# from . import standard

valfromroots_lnG = chebychev.valfromroots_lnG
coeff_canonicalization_gain = chebychev.coeff_canonicalization_gain


def roots(c, X_scale=1):
    c = np.asarray(c, complex)
    c2 = np.copy(c)
    if len(c) % 2 == 1:
        c2[1::2] *= 1j
    else:
        c2[0::2] *= -1j
    roots = chebychev.roots(c2, X_scale=X_scale)

    SGN = c[0] / c[-1]
    count = roots.imag < 0
    print(c)
    print("CHECK COUNT!", (c[0]), SGN, np.sum(count), len(roots))

    # for root in roots:
    #    print(root, abs(chebychev.val_lnG(root, c2, X_scale = X_scale)[0]))

    roots = 1j * roots

    return roots


def roots_lnG(c, X_scale=1):
    c = np.asarray(c, complex)
    c2 = np.copy(c)
    if len(c) % 2 == 1:
        c2[1::2] *= 1j
    else:
        c2[0::2] *= -1j
    roots, lnG = chebychev.roots_lnG(c2, X_scale=X_scale)
    roots = 1j * roots

    # TODO, repair the real poly conjugate matching
    return roots, lnG


def fromroots_lnG(roots, X_scale=1):
    roots = np.asarray(roots, complex)
    roots2 = roots * -1j
    c, lnG = chebychev.fromroots_lnG(roots2, X_scale=X_scale)
    if len(roots) % 2 == 0:
        c[1::2] *= -1j
    else:
        c[0::2] *= +1j
    assert np.all(c.imag == 0)
    return c.real, lnG


def val_lnG(X, c, X_scale=1, lnG=0):
    c2 = np.asarray(c, complex)
    c2 = np.copy(c2)
    c2[1::2] *= +1j
    if len(c2) % 2 == 1:
        # print('A2')
        return chebychev.val_lnG(X * -1j, c2, X_scale=X_scale, lnG=lnG)
    else:
        # print('B2')
        val, lnG = chebychev.val_lnG(X * -1j, c2, X_scale=X_scale, lnG=lnG)
        return -val, lnG


def vander_lnG(X, N, X_scale=1, lnG=0):
    V, lnG = chebychev.vander_lnG(X * +1j, N, X_scale=X_scale, lnG=lnG)
    V = V.astype(complex)
    if N % 2 == 1:
        V[:, 1::2] *= 1j
        V[:, 0::2] *= -1
    else:
        V[:, 1::2] *= -1j
    return V, lnG


def companion(c):
    c2 = np.asarray(c, complex)
    c2 = np.copy(c2)
    if len(c) % 2 == 1:
        c2[1::2] *= 1j
    else:
        c2[0::2] *= 1j
    return 1j * np.polynomial.chebyshev.chebcompanion(c2)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


import numpy as np

from ..codings_cmn import (
    CodingTypeZ,
    Ipi,
)


class CodingnlFnlA(CodingTypeZ):
    N_parameters = 2
    unstable = False
    p_nlF_Hz = 0
    p_nl_A = 0
    max_amplitude = 1

    def update(self, nlF_Hz, nl_A):
        self.p_nlF_Hz = nlF_Hz
        self.p_nl_A = nl_A

    def reduce(self):
        return [
            self.p_nlF_Hz,
            self.p_nl_A,
        ]

    def option_set(self, minimum_BW_Hz=None, **kwargs):
        super(CodingnlFnlA, self).option_set(**kwargs)
        if minimum_BW_Hz is not None:
            self.max_amplitude = 1 - (minimum_BW_Hz / self.sys.F_nyquist_Hz)
        # TODO, should modify p_nl_A for the new max_amplitude
        return

    def update_roots(self, r1):
        """ """
        # TODO should honor max_amplitude
        amp = abs(r1)
        F_Hz = np.angle(r1) / np.pi * self.sys.F_nyquist_Hz
        relF_Hz = F_Hz / self.sys.F_cutoff_Hz
        if relF_Hz >= 1:
            relF_Hz = 0.999
        elif relF_Hz <= -1:
            relF_Hz = 0.001
        self.p_nlF_Hz = relF_Hz / (1 - relF_Hz ** 2) ** 0.5

        if amp < 1:
            # amp = x / (1. + x**2)**.5
            # amp**2 = x**2 / (1. + x**2)
            # amp**2 (1. + x**2) = x**2
            # amp**2  = x**2 ( 1 - amp**2)
            # amp**2 / (1 - amp**2) = x**2
            self.unstable = False
            self.p_nl_A = amp / (1 - amp ** 2) ** 0.5
        else:
            # amp = (1. + x**2)**.5 / x
            # amp**2 =  (1. + x**2) / x**2
            # amp**2 x**2 = x**2 + 1
            # amp**2  = x**2 (amp**2 - 1)
            # amp**2 / (amp**2 - 1) = x**2
            self.unstable = True
            amp = 1 / amp
            self.p_nl_A = amp / (1 - amp ** 2) ** 0.5
        return

    @property
    def gain_effect(self):
        # always 1 since it is complex
        if not self.unstable:
            return 1
        else:
            return 1

    @property
    def relF_Hz(self):
        val = self.p_nlF_Hz / (1.0 + self.p_nlF_Hz ** 2) ** 0.5
        return val

    @property
    def amplitude(self):
        if not self.unstable:
            amp = self.max_amplitude * self.p_nl_A / (1.0 + self.p_nl_A ** 2) ** 0.5
        else:
            amp = (1.0 + self.p_nl_A ** 2) ** 0.5 / self.p_nl_A / self.max_amplitude
        return amp

    def transfer(self):
        # frequency, logarithmic amplitude
        amp = self.amplitude
        r = amp * np.cos(
            np.pi * self.sys.F_cutoff_Hz * self.relF_Hz / self.sys.F_nyquist_Hz
        )
        Xn = self.sys.Xzn_grid
        Xnsq = self.sys.Xzn_grid_sq
        return (amp * amp) * Xnsq - 2 * Xn * r + 1

    def derivative(self):
        if self.disable:
            return []
        # real/imaginary part of root
        if not self.unstable:
            sqp5 = (1.0 + self.p_nl_A ** 2) ** 0.5
            amp = self.max_amplitude * self.p_nl_A / sqp5
            DampDl = self.max_amplitude * (
                1 / sqp5 - self.p_nl_A ** 2 / (1.0 + self.p_nl_A ** 2) ** 1.5
            )
        else:
            sqp5 = (1.0 + self.p_nl_A ** 2) ** 0.5
            amp = sqp5 / self.p_nl_A / self.max_amplitude
            DampDl = (1 / sqp5 - sqp5 / self.p_nl_A ** 2) / self.max_amplitude

        Fsqp5 = (1.0 + self.p_nlF_Hz ** 2) ** 0.5
        relF_Hz = self.p_nlF_Hz / Fsqp5
        DrelFHzDp = 1 / Fsqp5 - self.p_nlF_Hz ** 2 / (1.0 + self.p_nlF_Hz ** 2) ** 1.5

        F_nyquist_Hz = self.sys.F_nyquist_Hz
        Xn = self.sys.Xzn_grid
        Xnsq = self.sys.Xzn_grid_sq
        rcos = np.cos(np.pi * self.sys.F_cutoff_Hz * relF_Hz / F_nyquist_Hz)
        rsin = np.sin(np.pi * self.sys.F_cutoff_Hz * relF_Hz / F_nyquist_Hz)
        return [
            DrelFHzDp
            * (2 * amp * np.pi * self.sys.F_cutoff_Hz / self.sys.F_nyquist_Hz)
            * (Xn * rsin),
            ((2 * amp * DampDl) * Xnsq - Xn * (2 * rcos * DampDl)),
        ]

    def roots_c(self):
        # real/imaginary part of root
        return [
            self.amplitude
            * np.exp(
                Ipi * abs(self.relF_Hz) * self.sys.F_cutoff_Hz / self.sys.F_nyquist_Hz
            )
        ]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


from ..codings_cmn import (
    CodingTypeZ,
)


class CodingRealBW(CodingTypeZ):
    N_parameters = 1
    unstable = False
    p_BW = 0
    restabilize = False

    def update(self, BW):
        self.p_BW = BW

    def reduce(self):
        return [self.p_BW]

    def update_roots(self, r1):
        assert r1.imag == 0
        if self.restabilize:
            if r1 > 1:
                r1 = 1 / r1
        self.p_BW = (1 - r1) * self.sys.F_nyquist_Hz
        return

    @property
    def F_Hz(self):
        return 0

    @property
    def BW(self):
        return self.p_BW

    def transfer(self):
        # real, linear BW
        amp = 1 - self.p_BW / self.sys.F_nyquist_Hz
        return 1 - amp * self.sys.Xzn_grid

    def derivative(self):
        return [
            self.sys.Xzn_grid / self.sys.F_nyquist_Hz,
        ]

    def roots_r(self):
        amp = 1 - self.p_BW / self.sys.F_nyquist_Hz
        return [amp]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


from wield.iirrational.v1 import (
    data2filter,
    FitAid,
)

from wield.iirrational.testing import IIRrational_data

# monkeypatch in the msurrogate annotations
FitAid._msurrogate_MT = False
data2filter._msurrogate_MT = True

__all__ = [
    "data2filter",
    "IIRrational_data",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


from .plots import IIRPlots

# REPLACE ME (if wanted)
plotter = IIRPlots()


def plot_fit(fitter, **kwargs):
    return plotter.plot_fit(fitter, **kwargs)


def plot_rel_comparison(fitter, fitter_ref, **kwargs):
    return plotter.plot_rel_comparison(fitter, fitter_ref, **kwargs)


def plot_ZP(fitter, **kwargs):
    return plotter.plot_ZP(fitter, **kwargs)


def plot_ZP(fitter, **kwargs):
    return plotter.plot_ZP(fitter, **kwargs)


def plot_ZP_S(fitter, **kwargs):
    return plotter.plot_ZP_S(fitter, **kwargs)


def plot_fitter_flag(fitter, **kwargs):
    return plotter.plot_fitter_flag(fitter, **kwargs)


def plot_fitter_flag_compare(fitter, fitter_ref, **kwargs):
    return plotter.plot_fitter_flag_compare(fitter, fitter_ref, **kwargs)


def plot_fitter_flag_residuals(fitter=None, **kwargs):
    return plotter.plot_fitter_flag_residuals(fitter, **kwargs)


def plot_ZP_grab(fitter, duals, **kwargs):
    return plotter.plot_ZP_grab(fitter, duals, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


from ..codings_cmn import (
    CodingTypeZ,
)


class CodingSOS(CodingTypeZ):
    N_parameters = 2
    p_c1 = 0
    p_c2 = 0
    # none means don't enforce mirroring, True false means mirror to respective side
    unstable = None

    def update(self, c1, c2):
        self.p_c1 = c1
        self.p_c2 = c2

    def reduce(self):
        return [self.p_c1, self.p_c2]

    def effective_params(self):
        if self.unstable is None:
            return self.p_c1, self.p_c2
        elif self.unstable:
            return self.p_c1, self.p_c2
        else:
            return self.p_c1, self.p_c2

    def transfer(self):
        # second order sections (2x roots either real or complex conj)
        c1, c2 = self.effective_params()
        Xn = self.sys.Xzn_grid
        Xnsq = self.sys.Xzn_grid_sq
        return c2 * Xnsq + Xn * c1 + 1

    def derivative(self):
        Xn = self.sys.Xzn_grid
        Xnsq = self.sys.Xzn_grid_sq
        return [Xn, Xnsq]

    def update_roots(self, r1, r2=None):
        """
        r2, may be unspecified, in which case it is assumed to be nothing, if r1 is real, or otherwise the conjugate of r1
        """
        if self.unstable is not None:
            if abs(r1) >= 1:
                self.unstable = True
            else:
                self.unstable = False

        if r2 is None:
            if r1.imag != 0:
                self.p_c2 = r1.real ** 2 + r1.imag ** 2
                self.p_c1 = -2 * r1.real
            else:
                self.p_c2 = 0
                self.p_c1 = -r1.real
        else:
            if r1.imag != 0:
                # TODO check conjugates
                self.p_c2 = r1.real ** 2 + r1.imag ** 2
                self.p_c1 = -2 * r1.real
            else:
                self.p_c2 = r1 * r2
                self.p_c1 = -(r1 + r2)
        return

    def roots(self):
        # second order sections (2x roots either real or complex conj)

        c1, c2 = self.effective_params()
        # a = c2, b = c1, c = 1
        disc = c1 * c1 - 4 * c2
        if disc >= 0:
            sqrt_disc = disc ** 0.5
            if c1 < 0:
                r1 = (-c1 + sqrt_disc) / 2
            else:
                r1 = (-c1 - sqrt_disc) / 2
            r2 = c2 / r1
            return [r1, r2]
        else:
            sqrt_disc = (-disc) ** 0.5
            r1 = (-c1 + sqrt_disc * 1j) / 2
            return [r1, r1.conjugate()]

    def roots_r(self):
        # second order sections (2x roots either real or complex conj)

        c1, c2 = self.effective_params()
        # a = c2, b = c1, c = 1
        disc = c1 * c1 - 4 * c2
        if disc >= 0:
            sqrt_disc = disc ** 0.5
            if c1 < 0:
                r1 = (-c1 + sqrt_disc) / 2
            else:
                r1 = (-c1 - sqrt_disc) / 2
            r2 = c2 / r1
            return [r1, r2]
        else:
            return []

    def roots_c(self):
        # second order sections (2x roots either real or complex conj)

        # complex form is always numerically stable
        c1, c2 = self.effective_params()
        # a = c2, b = c1, c = 1
        disc = c1 * c1 - 4 * c2
        if disc >= 0:
            sqrt_disc = disc ** 0.5
            return []
        else:
            sqrt_disc = (-disc) ** 0.5
            r1 = (-c1 + sqrt_disc * 1j) / 2
            return [r1]


class CodingSOSMirror(CodingSOS):
    unstable = False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


import numpy as np

from .base import (
    CodingType,
    Ipi,
    # I2pi
)
from ... import TFmath

# import scipy.linalg


class CodingResidues(CodingType):
    N_parameters = None

    def setup(self, coding_K, codings_r):
        self.coding_K = coding_K
        self.codings_r = codings_r
        self.codings_full = [coding_K, codings_r]

        N_parameters = 0
        for coding in self.codings_full:
            N_parameters += coding.N_parameters

        self.N_parameters = N_parameters
        return

    # def update_ZPK(self, ZPK):
    #    res_list, ZPKsum = TFmath.ZPK2residues()
    #    return

    def update(self, *params):
        idx_p = 0
        for coding in self.codings_full:
            subP = params[idx_p : idx_p + coding.N_parameters]
            coding.update(*subP)
            idx_p += coding.N_parameters
        return

    def reduce(self):
        params = []
        for coding in self.codings_full:
            params.extend(coding.reduce())
        return params

    def option_set(self, **kwargs):
        for coding in self.codings_full:
            coding.option_set(**kwargs)
        return

    def transfer(self, sys):
        val = 0
        for coding in self.codings_full:
            subT = coding.transfer(sys)
            val = val + subT
        return subT

    def derivative(self, sys):
        derivs = []
        for coding in self.codings_full:
            subD = coding.derivative(sys)
            derivs.extend(subD)
        return derivs

    def ZPK(self, sys):
        residue_list = []
        for c_r in self.codings_r:
            residue_list.extend(c_r.residues(sys))

        ZPKsum = self.coding_K.ZPK(sys)

        return TFmath.residues2ZPK(residue_list, ZPKsum)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""
from matplotlib import ticker
import numpy as np


def ticks_log_format(value, index):
    """
    get the value and returns the value as:
       integer: [0,99]
       1 digit float: [0.1, 0.99]
       n*10^m: otherwise
    To have all the number of the same size they are all returned as latex strings
    """
    exp = np.floor(np.log10(value))
    base = value / 10 ** exp
    if exp == 0 or exp == 1:
        return "${0:d}$".format(int(value))
    if exp == -1:
        return "${0:.1f}$".format(value)
    else:
        return "${0:d}\\times10^{{{1:d}}}$".format(int(base), int(exp))


def setup_log_xticks(ax):
    ax.xaxis.set_major_formatter(
        ticker.FuncFormatter(ticks_log_format)
    )  # add the custom ticks
    ax.xaxis.set_minor_formatter(
        ticker.FuncFormatter(ticks_log_format)
    )  # add the custom ticks



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""

from wield.bunch import DeepBunch
import os
import re
import errno
import os.path as path
import time


def create(
    doc_db,
    name,
    verbosity_limit=3,
):
    if doc_db is None:
        doc_db = DeepBunch()
        doc_db.verbosity_limit = verbosity_limit
    doc_db.name = name
    doc_db.time = time.time()
    doc_db.section = []
    return doc_db


def annotate(
    doc_db,
    name,
    verbosity,
    method,
    about,
    verbosity_limit=None,
    fitter=None,
    plotter=None,
):
    if name is None:
        doc = doc_db
        if doc is not None:
            doc.method = method
            doc.about = about
            doc.verbosity = verbosity
            doc.time = time.time()
            if not doc.verbosity:
                doc.verbosity = verbosity
            else:
                verbosity = doc.verbosity
            if doc.verbosity_limit:
                verbosity_limit = doc.verbosity_limit
            else:
                verbosity_limit = 10
            doc.plotter = plotter
            if verbosity_limit >= verbosity:
                if fitter is not None:
                    doc.fitter = fitter.copy()
                return doc
    else:
        if doc_db is not None:
            doc = doc_db[name]
            doc.name = name
            doc.method = method
            doc.about = about
            doc.time = time.time()
            if not doc.verbosity:
                doc.verbosity = verbosity
            else:
                verbosity = doc.verbosity
            if verbosity_limit is None:
                verbosity_limit = doc.verbosity_limit
            if not verbosity_limit:
                verbosity_limit = 10
            if not doc.verbosity_limit:
                doc.verbosity_limit = verbosity_limit
            else:
                verbosity_limit = doc.verbosity_limit
            doc.plotter = plotter
            if verbosity_limit >= verbosity:
                seq = doc_db.sequence
                if not seq:
                    seq = []
                    doc_db.sequence = seq
                doc.section = doc_db.section + [len(seq)]
                seq.append(name)
                if fitter is not None:
                    doc.fitter = fitter.__class__(copy=fitter)
                return doc


def annotate_into(
    doc_db,
    name,
):
    if doc_db is not None:
        seq = doc_db.sequence
        if not seq:
            seq = []
            doc_db.sequence = seq
        seq.append(name)
        doc = doc_db[name]
        doc.name = name
        doc.section = doc_db.section + [len(doc_db.sequence)]
        if doc_db.verbosity_limit:
            doc.verbosity_limit = doc_db.verbosity_limit
        return doc



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


import IPython

_ip = IPython.get_ipython()

if _ip is not None:
    _ip.magic("load_ext autoreload")
    _ip.magic("autoreload 2")

    # if this is run from the console then inline can't be found. This hack seems to get around it
    try:
        import ipykernel.pylab.backend_inline

        backend = ipykernel.pylab.backend_inline.InlineBackend.instance()
        backend.rc.clear()

        _ip.magic("matplotlib inline")
        _ip.magic("pylab inline")
    except Exception:
        _ip.magic("matplotlib")
        _ip.magic("pylab")
    # mpl.use('GTK3Cairo')

import numpy as np
import matplotlib as mpl
import matplotlib
from matplotlib import gridspec
import matplotlib.pyplot as plt

from IPython.display import (
    display,
    display_pretty,
    display_html,
    display_jpeg,
    display_png,
    display_json,
    display_latex,
    display_svg,
    display_javascript,
    display_markdown,
    FileLink,
    Image,
    SVG,
    clear_output,
    Audio,
    Javascript,
    Markdown,
)

# for more options in mpl
import wield.iirrational.utilities.mpl

from wield.iirrational.utilities.mpl.utils import (
    indexed_cmap,
)

from wield.iirrational.utilities.mpl import (
    AutoPlotSaver,
    mplfigB,
    asavefig,
    generate_stacked_plot_ax,
    style_6p5in,
    style_3p5in,
    style_tex_serif,
    style_serif,
    setup_log_xticks,
)


try:
    import tabulate
except ImportError:
    pass


def setup_auto_savefig(ipynb_name, check_warn=False):
    from os import path

    new = path.splitext(ipynb_name)[0] + "-ipynb"
    asavefig.org_subfolder = new


if _ip is not None:
    asavefig.fixname = True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""
import numpy as np
from wield.bunch import Bunch

from ..utilities import ensure_aid

from .. import fitters_ZPK
from .. import TFmath
from .. import plots
from .. import annotate
from . import algorithms


def Q_rank_calc(z, p):
    if p.real == 0 or z.real == 0:
        if p.real == z.real:
            Q_rank = 0
        else:
            # TODO
            # should use the data spacing to regularize this case
            Q_rank = 1e3
    else:
        Q_rank = abs(p - z) * (1 / (p.real) ** 2 + 1 / (z.real) ** 2) ** 0.5
    return Q_rank


def match_pairs(
    aid,
    duals=True,
):
    """
    Match unique closest pairs, if they are within a bandwidth of 0Hz, then they are ignored
    """
    aid = ensure_aid(aid)

    def coding_to_roots(coding_lst):
        coding_roots = []
        for coding in coding_lst:
            r_lst = coding.roots_c_Sf()
            if len(r_lst) == 0:
                coding_roots.append(None)
                continue
            elif len(r_lst) == 1:
                r = r_lst[0]
                if r.real > 0:
                    # TODO allow for unstable
                    # coding_roots.append(1/r[0].conjugate())
                    coding_roots.append(None)
                    continue
                else:
                    # then must be within 45deg
                    if r.imag < r.real:
                        # then this root is within a bandwidth of zero and is effectively a real root
                        # these require much more careful consideration
                        # TODO, check that the usage of roots_Sf is OK
                        coding_roots.append(None)
                        continue
                    coding_roots.append(r)
                    continue
            else:
                # TODO account for
                coding_roots.append(None)
                assert False
                continue
        return coding_roots

    coding_poles = coding_to_roots(aid.fitter.den_codings)
    coding_zeros = coding_to_roots(aid.fitter.num_codings)

    def nearest_idx(lst_1, lst_2):
        nearest_lst = []
        for r1 in lst_1:
            if r1 is None:
                nearest_lst.append(None)
                continue
            dist_nearest = float("inf")
            idx_nearest = None
            for idx_2, r2 in enumerate(lst_2):
                if r2 is None:
                    continue
                dist = abs(r1 - r2)
                if dist < dist_nearest:
                    idx_nearest = idx_2
                    dist_nearest = dist
            nearest_lst.append(idx_nearest)
        return nearest_lst

    z_nearest = nearest_idx(coding_zeros, coding_poles)
    p_nearest = nearest_idx(coding_poles, coding_zeros)

    if duals:
        # TODO, refactor the duals generation
        duals = []
        for idx_z, idx_p in enumerate(z_nearest):
            if idx_p is None:
                continue
            # coding_z = aid.fitter.num_codings[idx_z]
            # coding_p = aid.fitter.den_codings[idx_p]
            # TODO annotate about stability
            p = coding_poles[idx_p]
            z = coding_zeros[idx_z]
            Q_rank = Q_rank_calc(z, p)
            Q_rank_BW = (TFmath.abs_sq(p - z) + (p.real) ** 2 + (p.imag) ** 2) ** 0.5
            F_loc_Hz = (p.imag + z.imag) / 2
            if idx_z == p_nearest[idx_p]:
                duals.append(
                    Bunch(
                        idx_z=idx_z,
                        idx_p=idx_p,
                        Q_rank=Q_rank,
                        Q_rank_BW=Q_rank_BW,
                        F_loc_Hz=F_loc_Hz,
                        dist=abs(p - z),
                        z=z,
                        p=p,
                    )
                )
        duals.sort(key=lambda b: b.Q_rank)
        return duals
    else:
        pairs = []

        def loop(idx_z, idx_p):
            p = coding_poles[idx_p]
            z = coding_zeros[idx_z]
            Q_rank = Q_rank_calc(z, p)
            Q_rank_BW = (TFmath.abs_sq(p - z) + (p.real) ** 2 + (p.imag) ** 2) ** 0.5
            F_loc_Hz = (p.imag + z.imag) / 2
            pairs.append(
                Bunch(
                    idx_z=idx_z,
                    idx_p=idx_p,
                    Q_rank=Q_rank,
                    Q_rank_BW=Q_rank_BW,
                    F_loc_Hz=F_loc_Hz,
                    dist=abs(p - z),
                    z=z,
                    p=p,
                )
            )

        for idx_z, idx_p in enumerate(z_nearest):
            if idx_p is None:
                continue
            loop(idx_z, idx_p)
        for idx_p, idx_z in enumerate(p_nearest):
            if idx_z is None:
                continue
            if idx_p == z_nearest[idx_z]:
                # would be redundant
                continue
            loop(idx_z, idx_p)
        pairs.sort(key=lambda b: b.Q_rank)
        return pairs


def add_resrank(
    dual_b,
    aid,
    relupdown=2,
):
    aid = ensure_aid(aid)
    dual_b.idx_min = np.searchsorted(
        aid.fitter.F_Hz, dual_b.F_loc_Hz - relupdown * dual_b.Q_rank_BW
    )
    dual_b.idx_max = np.searchsorted(
        aid.fitter.F_Hz, dual_b.F_loc_Hz + relupdown * dual_b.Q_rank_BW
    )

    # aid.log(idx_min, idx_max)
    select = slice(dual_b.idx_min, dual_b.idx_max)
    # F_Hz = aid.fitter.F_Hz[select]
    Xsf = 1j * aid.fitter.F_Hz[select]
    W = aid.fitter.W[select]
    res = aid.fitter.residuals_preferred[select]
    h = TFmath.polyvalfromroots(Xsf, [dual_b.z]) / TFmath.polyvalfromroots(
        Xsf, [dual_b.p]
    )
    # aid.log(abs(h))
    res_h = W * (h.conjugate() ** -2 - h)
    R = aid.fitter.xfer_fit[select] / aid.fitter.data[select]
    res = W * ((R.conjugate()) ** -1 - R)
    R_mod = aid.fitter.xfer_fit[select] / h / aid.fitter.data[select]
    res_mod = W * (R_mod.conjugate() ** -1 - R_mod)

    dual_b.res_tot = np.sum(TFmath.abs_sq(res)) / (dual_b.idx_max - dual_b.idx_min)
    dual_b.res_h_tot = np.sum(TFmath.abs_sq(res_h)) / (dual_b.idx_max - dual_b.idx_min)
    dual_b.res_mod_tot = np.sum(TFmath.abs_sq(res_mod)) / (
        dual_b.idx_max - dual_b.idx_min
    )
    return dual_b


def order_reduce(
    aid,
    duals=None,
    Q_rank_cutoff=1,
    rolloff_order=0,
):
    aid = ensure_aid(aid)

    if duals is None:
        duals = match_pairs(aid.fitter)
    r1 = TFmath.abs_sq(aid.fitter.residuals.resP)

    num_codings = []
    den_codings = []
    exclude_p = set()
    exclude_z = set()
    duals_used = []
    duals_rejected = []
    for b in duals:
        used = False
        if b.Q_rank < Q_rank_cutoff:
            exclude_p.add(b.idx_p)
            exclude_z.add(b.idx_z)
            duals_used.append(b)
            used = True
        else:
            F_idx = np.searchsorted(aid.fitter.F_Hz, b.F_loc_Hz)
            F_idx_low = max(0, F_idx - 2)
            F_idx_high = min(len(aid.fitter.F_Hz) - 1, F_idx + 2)
            # get the local mean width of bins to restrict bandwidths
            F_mean_Hz = (aid.fitter.F_Hz[F_idx_high] - aid.fitter.F_Hz[F_idx_low]) / (
                F_idx_high - F_idx_low
            )
            if b.Q_rank_BW < F_mean_Hz / 2:
                exclude_p.add(b.idx_p)
                exclude_z.add(b.idx_z)
                duals_used.append(b)
                aid.log(F_mean_Hz, b.Q_rank_BW)
                aid.log("Removing by bandwidth")
                used = True

        if not used:
            duals_rejected.append(b)

    # switch method to a residual ranking
    duals.sort(key=lambda b: b.F_loc_Hz)
    if False and not duals_used:
        duals_rejected = []
        for b in duals:
            used = False
            add_resrank(aid.fitter, b)
            if b.idx_max - b.idx_min < 2:
                # kill single point pairs (they shouldn't exist anyway)
                exclude_p.add(b.idx_p)
                exclude_z.add(b.idx_z)
                duals_used.append(b)
                used = True

            if used:
                pass
            if not used:
                duals_rejected.append(b)

    for idx_p, coding in enumerate(aid.fitter.den_codings):
        if idx_p not in exclude_p:
            den_codings.append(coding)
    for idx_z, coding in enumerate(aid.fitter.num_codings):
        if idx_z not in exclude_z:
            num_codings.append(coding)

    fitter_out = aid.fitter.__class__(
        parent=aid.fitter,
        num_codings=num_codings,
        den_codings=den_codings,
    )

    return Bunch(locals())


def order_reduce_weakroots(
    aid,
    delay_s,
):
    """ """

    aid = ensure_aid(aid)

    F_largest_Hz = np.max(aid.fitter.F_Hz)
    Xsf = 1j * F_largest_Hz

    def ang_mag_mapping(coding_list, rtype):
        new_list = []
        for coding in coding_list:
            try:
                roots = coding.roots_Sf()
            except fitters_ZPK.BranchCutAmbiguity:
                continue
            if np.any(abs(np.real(roots)) < 1 * F_largest_Hz):
                new_list.append(None)
                continue
            # aid.log(roots, rtype)
            # divide by the distance to the roots to normalize DC to 1
            TF = TFmath.polyvalfromroots(Xsf, roots) / np.prod(np.abs(roots))
            mag = abs(TF)
            angle = TF / mag * np.prod(-np.sign(roots))
            # print(roots, mag, angle)
            new_list.append((angle, np.log(mag)))
        # the LAST entry is the null-map, so it must be checked-for
        new_list.append((1, 0))
        return new_list

    z_map = ang_mag_mapping(aid.fitter.num_codings, "zeros")
    p_map = ang_mag_mapping(aid.fitter.den_codings, "poles")
    # print(z_map, p_map)

    def nearest_idx(lst_1, lst_2, num_first=True):
        nearest_lst = []
        for idx_1, pair_1 in enumerate(lst_1):
            if pair_1 is None:
                continue
            ang_1, mag_1 = pair_1
            dist_nearest = float("inf")
            ang_diff_nearest = float("inf")
            mag_diff_nearest = float("inf")
            idx_nearest = None
            for idx_2, pair_2 in enumerate(lst_2):
                if pair_2 is None:
                    continue
                if (idx_2 == len(lst_2) - 1) and (idx_1 == len(lst_1) - 1):
                    continue
                ang_2, mag_2 = pair_2
                ang_diff = np.angle(ang_1 * ang_2.conjugate())
                mag_diff = mag_1 - mag_2
                dist = (ang_diff ** 2 + 1 * mag_diff ** 2) ** 0.5
                if dist < dist_nearest:
                    idx_nearest = idx_2
                    dist_nearest = dist
                    ang_diff_nearest = ang_diff
                    mag_diff_nearest = mag_diff
            if idx_nearest is None:
                continue
            if idx_1 < len(lst_1) - 1:
                if num_first:
                    coding_num = aid.fitter.num_codings[idx_1]
                    idx_num = idx_1
                else:
                    coding_den = aid.fitter.den_codings[idx_1]
                    idx_den = idx_1
            else:
                if num_first:
                    coding_num = None
                    idx_num = None
                else:
                    coding_den = None
                    idx_den = None
            if idx_nearest < len(lst_2) - 1:
                if not num_first:
                    coding_num = aid.fitter.num_codings[idx_nearest]
                    idx_num = idx_nearest
                else:
                    coding_den = aid.fitter.den_codings[idx_nearest]
                    idx_den = idx_nearest
            else:
                if not num_first:
                    coding_num = None
                    idx_num = None
                else:
                    coding_den = None
                    idx_den = None
            data = Bunch(
                idx_nearest=idx_nearest,
                coding_num=coding_num,
                coding_den=coding_den,
                idx_num=idx_num,
                idx_den=idx_den,
                dist=dist_nearest,
                ang_diff=ang_diff_nearest,
                mag_diff=mag_diff_nearest,
            )
            nearest_lst.append(data)
        return nearest_lst

    z_nearest = nearest_idx(z_map, p_map, num_first=True)
    p_nearest = nearest_idx(p_map, z_map, num_first=False)

    key = lambda b: b.dist if b is not None else float("inf")
    if z_nearest:
        z_min = min(z_nearest, key=key)
    else:
        z_min = None
    if p_nearest:
        p_min = min(p_nearest, key=key)
    else:
        p_min = None
    if p_min is None:
        if z_min is None:
            return False
        else:
            zp_min = z_min
    else:
        if z_min is None:
            zp_min = p_min
        else:
            zp_min = min(z_min, p_min, key=key)

    num_codings = list(aid.fitter.num_codings)
    den_codings = list(aid.fitter.den_codings)
    if zp_min.idx_num is not None:
        del num_codings[zp_min.idx_num]
    if zp_min.idx_den is not None:
        del den_codings[zp_min.idx_den]

    if zp_min.coding_num is not None:
        pr_num = zp_min.coding_num.roots_Sf()
    else:
        pr_num = None

    if zp_min.coding_den is not None:
        pr_den = zp_min.coding_den.roots_Sf()
    else:
        pr_den = None

    aid.log("WEAK REMOVE: ", pr_num, pr_den)
    fitter_new = aid.fitter.copy(
        num_codings=num_codings,
        den_codings=den_codings,
    )
    ratio = fitter_new.residuals_average / aid.fitter.residuals_average
    aid.log("RATIO for WEAK1: ", ratio)
    # aid.log("BEFORE fitters_ZPK: ", aid.fitter.residuals_average)
    # aid.log("BEFORE NEW: ", fitter_new.residuals_average)

    # maybe feed forward the delay
    # fitter_new.optimize_delay(only_delay = False)
    if delay_s is None:
        fitter_new.optimize_delay(
            residuals_type=aid.hint("residuals_type_alt", default="log"),
            only_delay=False,
            rescale=True,
            aid=aid,
        )
        fitter_new.optimize_delay(
            rescale=True,
            only_delay=False,
            aid=aid,
        )
    else:
        fitter_new.optimize(
            residuals_type=aid.hint("residuals_type_alt", default="log"),
            rescale=True,
            aid=aid,
        )
        fitter_new.optimize(
            rescale=True,
            aid=aid,
        )

    ratio = fitter_new.residuals_average / aid.fitter.residuals_average
    aid.log("RATIO for WEAK2: ", ratio)
    algorithms.sign_check_flip(fitter_new)
    return aid.fitter_check(
        fitter_new,
        hint_name="weakroots",
        variant="OrdDn",
    )


def order_reduce_aggressive(aid):
    aid = ensure_aid(aid)

    N = 1
    # DOESN'T SEEM TO HELP
    # duals = match_pairs(aid.fitter)
    # ignore_orig = aid.fitter.codings_ignore
    # for dualb in duals:
    #    aid.fitter.codings_ignore = aid.fitter.root_codings_set - set([
    #        aid.fitter.num_codings[dualb.idx_z],
    #        aid.fitter.den_codings[dualb.idx_p]
    #    ]).union(ignore_orig)
    #    aid.fitter.optimize()
    # aid.fitter.codings_ignore = ignore_orig
    while True:
        duals = match_pairs(aid=aid)
        # TODO annotate better
        # at this cutoff, the pairs are essentially guaranteed not to affect the residual in a meaningful way
        ORb = order_reduce(
            aid=aid,
            duals=duals,
            Q_rank_cutoff=0.7,
        )

        def closure(ORb):
            def plot_duals(fitter):
                axB = plots.plot_fit(fitter)
                plots.plot_ZP_grab(fitter, ORb.duals_used, axB=axB, color="black")
                plots.plot_ZP_grab(fitter, ORb.duals_rejected, axB=axB, color="blue")
                return axB

            return plot_duals

        aid.annotate(
            name="duals_ID_{0}".format(N),
            method="MultiReprFilter.optimize",
            fitter=aid.fitter,
            plotter=closure(ORb),
            verbosity=4,
            about=(
                """
                ID Pairs for order reduction
                """
            ),
        )
        if ORb.duals_used:
            aid.fitter_update(ORb.fitter_out)

        further_improvement = False
        if not ORb.duals_used:
            duals = match_pairs(duals=False, aid=aid)
            duals = [add_resrank(b, aid=aid) for b in duals]
            duals.sort(key=lambda b: b.res_h_tot)
            for b_low in duals[:4]:
                # aid.log(b_low.res_h_tot, b_low.res_mod_tot, b_low.res_tot)
                num_codings = list(aid.fitter.num_codings)
                den_codings = list(aid.fitter.den_codings)
                # aid.log("LEN:", len(aid.fitter.num_codings), len(aid.fitter.den_codings))
                # aid.log("IDX:", b_low.idx_z, b_low.idx_p)
                del num_codings[b_low.idx_z]
                del den_codings[b_low.idx_p]

                fitter_new = aid.fitter.__class__(
                    parent=aid.fitter,
                    num_codings=num_codings,
                    den_codings=den_codings,
                )
                fitter_new.optimize(aid=aid)
                res_ratio = fitter_new.residuals_average / aid.fitter.residuals_average
                aid.log("RATIO: ", res_ratio)
                if res_ratio < 1.0:
                    aid.log("fit improved from pair at", b_low.F_loc_Hz)
                    further_improvement = True
                    aid.fitter_update(fitter_new)
                    break
                else:
                    aid.log("fit NOT improved from pair at", b_low.F_loc_Hz)
        if not ORb.duals_used and not further_improvement:
            break

        aid.fitter.optimize(aid=aid)
        # annotate.annotate(
        #    doc_db,
        #    name = 'seq_order_reduce_opt_{0}'.format(N),
        #    method = 'MultiReprFilter.optimize',
        #    fitter   = aid.fitter,
        #    plotter  = plots.plot_fit,
        #    verbosity = 6,
        #    about = (
        #        """
        #        Optimize again after order reduction
        #        """
        #    ),
        # )
        N += 1
        aid.log("N: ", N)
    return


def order_reduce_negroots(aid):
    """
    ONLY works on F_nyquist_Hz > 2 * F_max_Hz
    """
    aid = ensure_aid(aid)

    # Only works in Z domain
    assert aid.fitter.F_nyquist_Hz is not None

    # F_largest_Hz = np.max(aid.fitter.F_Hz)
    def BW_check(coding_list, rtype):
        new_list = []
        for coding in coding_list:
            root_list = np.asarray(coding.roots())
            # exclude positive ones
            if np.any(root_list.real > 0):
                new_list.append(coding)
                continue
            aid.log("REMOVING ", rtype, root_list)
        return new_list

    num_codings = BW_check(aid.fitter.num_codings, "Z: ")
    den_codings = BW_check(aid.fitter.den_codings, "P: ")

    fitter_new = aid.fitter.__class__(
        parent=aid.fitter,
        num_codings=num_codings,
        den_codings=den_codings,
    )
    aid.fitter_update(fitter_new)
    return



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


import numpy as np
import matplotlib.collections as mplcollect

# import matplotlib as mpl
# from matplotlib import patches
import matplotlib.transforms as mpltrans

from . import utilities


def plot_ZP_grab(
    self,
    fitter,
    duals,
    color="black",
    axB=None,
):
    def rlog_F(r):
        F_Hz = r.imag
        BW = abs(r.real)
        # select = abs(r) < 1
        return BW, F_Hz

    ax = axB.mag2
    x_z = []
    y_z = []
    x_p = []
    y_p = []
    for b in duals:
        # coding_z = fitter.num_codings[b.idx_z]
        # z_rl, z_F = rlog_F(coding_z.roots_c(fitter)[0])
        z_rl, z_F = rlog_F(b.z)
        # coding_p = fitter.den_codings[b.idx_p]
        # p_rl, p_F = rlog_F(coding_p.roots_c(fitter)[0])
        p_rl, p_F = rlog_F(b.p)

        trans = mpltrans.composite_transform_factory(
            ax.transData,
            ax.transAxes.inverted(),
        )
        locs_z = trans.transform((z_F, z_rl))
        locs_p = trans.transform((p_F, p_rl))
        x_z.append(locs_z[0])
        y_z.append(locs_z[1])
        x_p.append(locs_p[0])
        y_p.append(locs_p[1])
    x_z = np.asarray(x_z)
    y_z = np.asarray(y_z)
    x_p = np.asarray(x_p)
    y_p = np.asarray(y_p)

    aspect = utilities.get_aspect(ax)
    r = (0.01 + ((x_z - x_p) ** 2 + (y_z - y_p) ** 2 / aspect ** 1) ** 0.5,)
    angles = np.angle((x_z - x_p) + 1j * (y_z - y_p) / aspect, deg=True)
    col = mplcollect.EllipseCollection(
        widths=r,
        heights=0.012,
        angles=angles,
        units="width",
        facecolors="none",
        edgecolors=color,
        offsets=np.vstack([(x_z + x_p) / 2, (y_z + y_p) / 2]).T,
        transOffset=ax.transAxes,
    )
    ax.add_collection(col)
    return axB



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


from ... import TFmath

from ...utilities import ensure_aid

from . import algorithms


def skip_list_listing(group, non_mindelay=True):
    zr_z = []
    zr_i = []
    for idx_z, z in enumerate(group):
        if z.real < 0 and not non_mindelay:
            continue
        zr_z.append(z)
        zr_i.append(idx_z)
    # print(zr_i, zr_z)
    return zip(zr_i, zr_z)


def ranking_delay_flip(
    aid,
    marginalize_delay=False,
    non_mindelay=True,
    representative=False,
    hint_name=None,
):
    aid = ensure_aid(aid)
    rank_zp_idx_list = []
    best_worst_list = []

    Zr = aid.fitter.zeros.r
    Zc = aid.fitter.zeros.c

    for idx_z, z in skip_list_listing(Zr, non_mindelay=False):
        z_flip = -z.conjugate()
        pb = algorithms.resrank_program(
            aid.fitter,
            rank_zp_idx_list,
            "flip_r",
            ["ZrDelIdx", idx_z, "ZrAdd", z_flip],
            variant="OrdFl",
            marginalize_delay=marginalize_delay,
            idx=("r", idx_z),
        )
        rank = pb.rank
        best_worst_list.append([rank, "r", idx_z, z, pb])

    for idx_z, z in skip_list_listing(Zc, non_mindelay):
        z_flip = -z.conjugate()
        pb = algorithms.resrank_program(
            aid.fitter,
            rank_zp_idx_list,
            "flip_c",
            ["ZcDelIdx", idx_z, "ZcAdd", z_flip],
            variant="OrdFl",
            marginalize_delay=marginalize_delay,
            idx=("c", idx_z),
        )
        rank = pb.rank
        best_worst_list.append([rank, "c", idx_z, z, pb])

    if not best_worst_list:
        return None

    best_worst_list.sort()
    best_worst_list.reverse()

    N = 0
    while best_worst_list:
        if N > 3:
            break
        bw = best_worst_list.pop()
        rank = bw[0]
        rank_zp_idx_list = bw[4:]
        trials = algorithms.ranking_reduction_trials(
            aid,
            rank_zp_idx_list,
            greedy=True,
            num_total_max=4,
            hint_name=hint_name,
        )
        trial = trials[0]
        did_reduce = aid.fitter_check(
            trial.fitter,
            variant="OrdC",
            hint_name=hint_name,
        )
        if did_reduce:
            aid.fitter_update(
                trial.fitter,
                representative=representative,
            )
            if not representative:
                aid.fitter_checkpoint()
            aid.log_progress(
                5,
                ("zero flipped, maxzp {}, residuals={:.2e}, reldeg={}").format(
                    aid.fitter_orders().maxzp, aid.fitter.residuals_average, aid.fitter_orders().reldeg,
                ),
            )
            break
        N += 1

    return did_reduce


def order_reduce_flip(
        aid,
        non_mindelay=True,
        representative=True,
        hint_name=None,
):
    #HACK, needs to happen or flip get stuck in a loop
    hint_name = None
    aid.log_progress(
        5,
        ("zero flipping, maxzp {}, residuals={:.2e}, {:.2e}, reldeg={}").format(
            aid.fitter_orders().maxzp, aid._fitter_current.residuals_average, aid.fitter.residuals_average, aid.fitter_orders().reldeg,
        ),
    )
    # import pdb; pdb.set_trace()

    check_sign = aid.fitter.check_sign
    for i in range(30):
        ret = ranking_delay_flip(
            aid,
            marginalize_delay=False,
            non_mindelay=non_mindelay,
            representative=representative,
            hint_name=hint_name,
        )
        if ret is None:
            break
        if not ret:
            break
    aid.fitter.check_sign = check_sign
    return


def order_reduce_flip2(
        aid,
        non_mindelay=True,
        representative=True,
        hint_name=None,
):
    aid.log_progress(
        5,
        ("zero flipping, maxzp {}, residuals={:.2e}, {:.2e}, reldeg={}").format(
            aid.fitter_orders().maxzp, aid._fitter_current.residuals_average, aid.fitter.residuals_average, aid.fitter_orders().reldeg,
        ),
    )

    from ... import fitters_ZPK
    fitter_new = aid.fitter.regenerate(
        ZPKrep=aid.fitter.ZPKrep,
        coding_map=fitters_ZPK.codings_s.coding_maps.SOSnlpoles,
        check_sign=False,
    )
    with fitter_new.with_codings_only([fitter_new.gain_coding]):
        fitter_new.optimize()
    fitter_new.optimize()
    algorithms.sign_check_flip(fitter_new)
    aid.fitter_update(fitter_new)

    fitter_new = aid.fitter.regenerate(
        ZPKrep=fitter_new.ZPKrep,
        # coding_map=fitters_ZPK.codings_s.coding_maps.SOSnlpoles,
        # check_sign=True,
    )
    aid.fitter_update(
        fitter_new,
    )
    aid.fitter_checkpoint()
    aid.log_progress(
        5,
        ("zero flipping, maxzp {}, residuals={:.2e}, {:.2e}, reldeg={}").format(
            aid.fitter_orders().maxzp, aid._fitter_current.residuals_average, aid.fitter.residuals_average, aid.fitter_orders().reldeg,
        ),
    )
    return



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
""" Setup and build the wield.collection package, which acts as the metapackage

Packaging guidance may be found at https://packaging.python.org/tutorials/packaging-projects/
"""
from setuptools import setup

# Settings are primarily in setup.cfg
setup(
    # this ensures setuptools is new enough to use setup.cfg
    setup_requires=[
        "setuptools >= 45.0.0",
    ],
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


from ..codings_cmn import (
    CodingTypeZ,
)


class CodingRealnlA(CodingTypeZ):
    N_parameters = 1
    unstable = False
    p_nl_A = 0
    max_amplitude = 1

    def update(self, nl_A):
        self.p_nl_A = nl_A

    def reduce(self):
        return [self.p_nl_A]

    def update_roots(self, r1):
        assert r1.imag == 0
        amp = r1

        if abs(amp) < 1:
            # amp = x / (1. + x**2)**.5
            # amp**2 = x**2 / (1. + x**2)
            # amp**2 (1. + x**2) = x**2
            # amp**2  = x**2 ( 1 - amp**2)
            # amp**2 / (1 - amp**2) = x**2
            self.unstable = False
            self.p_nl_A = amp / (1 - amp ** 2) ** 0.5
        else:
            # amp = (1. + x**2)**.5 / x
            # amp**2 =  (1. + x**2) / x**2
            # amp**2 x**2 = x**2 + 1
            # amp**2  = x**2 (amp**2 - 1)
            # amp**2 / (amp**2 - 1) = x**2
            self.unstable = True
            amp = 1 / amp
            self.p_nl_A = amp / (1 - amp ** 2) ** 0.5
        return

    def option_set(self, minimum_BW_Hz=None, **kwargs):
        super(CodingRealnlA, self).option_set(**kwargs)
        if minimum_BW_Hz is not None:
            self.max_amplitude = 1 - (minimum_BW_Hz / self.sys.F_nyquist_Hz)
        # TODO, should modify p_nl_A for the new max_amplitude
        return

    @property
    def amplitude(self):
        if not self.unstable:
            amp = self.max_amplitude * self.p_nl_A / (1.0 + self.p_nl_A ** 2) ** 0.5
        else:
            amp = (1.0 + self.p_nl_A ** 2) ** 0.5 / (self.p_nl_A * self.max_amplitude)
        return amp

    @property
    def gain_effect(self):
        if self.unstable:
            return -1
        else:
            return 1

    @property
    def F_Hz(self):
        return 0

    def transfer(self):
        # real, log amplitude, positive
        return 1 - self.amplitude * self.sys.Xzn_grid

    def derivative(self):
        if self.disable:
            return []
        # real/imaginary part of root
        if not self.unstable:
            sqp5 = (1.0 + self.p_nl_A ** 2) ** 0.5
            # amp = self.p_nl_A / sqp5
            DampDl = (
                1 / sqp5 - self.p_nl_A ** 2 / (1.0 + self.p_nl_A ** 2) ** 1.5
            ) * self.max_amplitude
        else:
            sqp5 = (1.0 + self.p_nl_A ** 2) ** 0.5
            # amp = sqp5 / self.p_nl_A
            DampDl = (1 / sqp5 - sqp5 / self.p_nl_A ** 2) / self.max_amplitude

        return [
            -DampDl * self.sys.Xzn_grid,
        ]

    def roots_r(self):
        return [self.amplitude]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""
import numpy as np
from wield.bunch import Bunch

from ..utilities import args

from .utilities import ZPKrep2Sf
from ..utilities.np import domain_sort
from .. import fitters_ZPK


def angle_cutter(
    F_Hz,
    data,
    deg=False,
    semiroll=False,
):
    F_Hz, data = domain_sort(F_Hz, data)
    ang = np.angle(data)

    above = ang > np.pi / 2
    below = ang < -np.pi / 2
    jumps = (above[:-1] & below[1:]) | (below[:-1] & above[1:])

    jump_locs = np.nonzero(jumps)[0]

    if semiroll:
        for idx1, idx2 in zip(jump_locs[:-1] + 1, jump_locs[1:]):
            if np.all(above[idx1:idx2]):
                ang[idx1:idx2] -= np.pi
            elif np.all(below[idx1:idx2]):
                ang[idx1:idx2] += np.pi
            else:
                # it is crossing, do nothing
                pass

    ins_locs = jump_locs + 1
    ins_F_Hz = (F_Hz[jump_locs] + F_Hz[jump_locs]) / 2
    F_Hz = np.insert(F_Hz, ins_locs, ins_F_Hz)
    if deg:
        ang = ang * 180 / np.pi
    ang = np.insert(ang, ins_locs, np.nan)
    return (F_Hz, ang)


def plot_augment_data(
    self,
    ax_mag=None,
    ax_phase=None,
    data_list=[],
    with_error=True,
    label=args.UNSPEC,
    markersize=3,
    marker=".",
    linestyle="",
    residuals=False,
    **kwargs
):
    label_default = args.argscan(label, self.label_data, "data")

    def plot(fitter, label=None, db_ref=None, color=None, **kwargs):
        if label is None:
            label = label_default
        db = Bunch()
        db.rep = fitter

        if color is None and db_ref is not None:
            color = db_ref.mline.get_color()

        if marker is not None:
            kwargs["marker"] = marker
        if markersize is not None:
            kwargs["markersize"] = markersize
        if linestyle is not None:
            kwargs["linestyle"] = linestyle

        f = fitter.F_Hz
        if not residuals:
            h = fitter.data
        else:
            h = fitter.data / fitter.xfer_fit
        if with_error:
            W = 0.001 + fitter.W
            select = f > 0
            if not residuals:
                db.min_data_err_lim = np.min(abs(h)) * max(1 / 3, np.min(1 / (1 + 1 / W)))
                db.max_data_err_lim = np.max(abs(h)) * min(3, np.max((1 + 1 / W)))
            else:
                db.min_data_err_lim = np.min(abs(h)) / 1.01
                db.max_data_err_lim = np.max(abs(h)) * 1.01
            if ax_mag:
                # ax_mag.fill_between(
                #    fitter.F_Hz[select],
                #    np.maximum((abs(h) / (1 + 1 / W))[select], np.min(abs(h)) / 100),
                #    np.minimum((abs(h) * (1 + 1 / W))[select], np.max(abs(h)) * 100),
                #    color = color,
                #    alpha = .2,
                #    **kwargs
                # )
                abs_h = abs(h)
                ax_mag.errorbar(
                    fitter.F_Hz[select],
                    abs_h[select],
                    yerr=(
                        abs_h[select]
                        - np.maximum(
                            (abs_h / (1 + 1 / W))[select], np.min(abs_h[select]) / 100
                        ),
                        np.minimum(
                            (abs_h * (1 + 1 / W))[select], np.max(abs_h[select]) * 100
                        )
                        - abs_h[select],
                    ),
                    color=color,
                    label=label,
                    **kwargs
                )
            if ax_phase:
                angle_h = np.angle(h, deg=True)
                ax_phase.errorbar(
                    f[select],
                    angle_h[select],
                    (
                        angle_h[select]
                        - np.maximum((angle_h - 180 / np.pi / W)[select], -180),
                        np.minimum((angle_h + 180 / np.pi / W)[select], 180)
                        - angle_h[select],
                    ),
                    color=color,
                    label=label,
                    **kwargs
                )
                # ax_phase.fill_between(
                #    f[select],
                #    np.minimum((np.angle(h, deg = True) + 180/np.pi / W)[select], 180),
                #    np.maximum((np.angle(h, deg = True) - 180/np.pi / W)[select], -180),
                #    color = color,
                #    alpha = .2,
                #    **kwargs
                # )
        else:
            if ax_mag:
                db.mline = ax_mag.plot(
                    fitter.F_Hz, abs(h), label=label, **kwargs
                )[0]

                if color is None:
                    color = db.mline.get_color()

            if ax_phase:
                _f, _h = angle_cutter(fitter.F_Hz, h, deg=True)
                db.pline = ax_phase.semilogx(
                    _f, _h, label=label, color=color, **kwargs
                )[0]

            if not residuals:
                db.min_data_err_lim = np.min(abs(h)) / 1.2
                db.max_data_err_lim = np.max(abs(h)) * 1.2
            else:
                db.min_data_err_lim = np.min(abs(h)) / 1.01
                db.max_data_err_lim = np.max(abs(h)) * 1.01

        midpt = (np.max(fitter.F_Hz) - np.min(fitter.F_Hz)) / 2
        if np.count_nonzero(fitter.F_Hz < midpt) < 0.65 * len(fitter.F_Hz):
            db.scale_log = False
        else:
            db.scale_log = True

        db.maxF_Hz = max(f)
        db.minF_Hz = min(f)
        select = f > 0
        db.minF_HzNZ = min(f[select])
        if not residuals:
            ax_mag.set_yscale('log_zoom')
        else:
            ax_mag.set_yscale('log_zoom')

        data_list.append(db)
        return db

    return plot


def plot_augment_fit(
    self,
    ax_mag,
    ax_phase,
    fit_list,
    with_error=False,
    F_Hz=None,
    residuals=False,
    label=args.UNSPEC,
):
    label_default = args.argscan(
        label, self.label_fit, "fit ({nP}P, {nZ}Z, Rsq ={Rsq:.3e})"
    )

    def plot(fitter, label=None, db_ref=None, color=None, **kwargs):
        if label is None:
            label = label_default
        label = label.format(
            nP=len(fitter.ZPKrep.poles),
            nZ=len(fitter.ZPKrep.zeros),
            Rsq=fitter.residuals_average,
        )

        if color is None and db_ref is not None:
            color = db_ref.mline.get_color()

        if F_Hz is not None:
            f = F_Hz
            h = fitter.xfer_eval(f)
        else:
            h = fitter.xfer_fit
            f = fitter.F_Hz
        if residuals:
            h[:] = 1
        db = Bunch()

        db.mline = ax_mag.plot(f, abs(h), label=label, color=color, **kwargs)[0]

        if color is None:
            color = db.mline.get_color()

        if ax_phase:

            _f, _h = angle_cutter(f, h, deg=True)
            db.pline = ax_phase.semilogx(_f, _h, label=label, color=color, **kwargs)

        if with_error and isinstance(fitter, fitters_ZPK.MultiReprFilterZ):
            select = f > 0
            ax_mag.fill_between(
                f[select],
                np.maximum(
                    (abs(h) / (1 + fitter.xfer_fit_error.mag_rel2))[select],
                    np.min(abs(h)) / 10,
                ),
                np.minimum(
                    (abs(h) * (1 + fitter.xfer_fit_error.mag_rel2))[select],
                    np.max(abs(h)) * 10,
                ),
                color=db.mline.get_color(),
                alpha=0.3,
            )
            if ax_phase:
                ax_phase.fill_between(
                    f[select],
                    np.minimum(
                        (
                            np.angle(h, deg=True)
                            + 180 / np.pi * fitter.xfer_fit_error.phase
                        )[select],
                        180,
                    ),
                    np.maximum(
                        (
                            np.angle(h, deg=True)
                            - 180 / np.pi * fitter.xfer_fit_error.phase
                        )[select],
                        -180,
                    ),
                    color=db.pline[0].get_color(),
                    alpha=0.3,
                )

                if not residuals:
                    db.min_fit_err_lim = np.min(abs(h)) * max(
                        1 / 3, np.min(1 / (1 + fitter.xfer_fit_error.mag_rel2)[select])
                    )
                    db.max_fit_err_lim = np.max(abs(h)) * min(
                        3, np.min((1 + fitter.xfer_fit_error.mag_rel2)[select])
                    )
                else:
                    db.min_fit_err_lim = np.min(abs(h)) / 1.01
                    db.max_fit_err_lim = np.max(abs(h)) * 1.01
        else:
            if not residuals:
                db.min_fit_err_lim = np.min(abs(h)) / 1.2
                db.max_fit_err_lim = np.max(abs(h)) * 1.2
            else:
                db.min_fit_err_lim = np.min(abs(h)) / 1.01
                db.max_fit_err_lim = np.max(abs(h)) * 1.01
        db.median_fit_lim = np.median(abs(h))

        midpt = (np.max(fitter.F_Hz) - np.min(fitter.F_Hz)) / 2
        if np.count_nonzero(fitter.F_Hz < midpt) < 0.65 * len(fitter.F_Hz):
            db.scale_log = False
        else:
            db.scale_log = True

        db.maxF_Hz = max(f)
        db.minF_Hz = min(f)
        select = f > 0
        db.minF_HzNZ = min(f[select])

        fit_list.append(db)
        return db

    return plot


def plot_augment_residuals(
    self,
    ax_mag,
    ax_phase,
    fit_list,
    label=args.UNSPEC,
):
    label_default = args.argscan(
        label, self.label_fit, "fit ({nP}P, {nZ}Z, Rsq ={Rsq:.3e})"
    )

    def plot(
        fitter, label=None, db_ref=None, color=None, markersize=3, marker=".", **kwargs
    ):
        if label is None:
            label = label_default
        label = label.format(
            nP=len(fitter.ZPKrep.poles),
            nZ=len(fitter.ZPKrep.zeros),
            Rsq=fitter.residuals_average,
        )

        if color is None and db_ref is not None:
            color = db_ref.mline.get_color()

        r = fitter.residuals_preferred
        f = fitter.F_Hz
        db = Bunch()

        if markersize is not None:
            kwargs["s"] = markersize
        db.mline = ax_mag.scatter(
            f, abs(r), label=label, color=color, marker=marker, **kwargs
        )
        ax_mag.set_yscale("log_zoom")

        if color is None:
            color = db.mline.get_color()

        if ax_phase:
            _f, _r = angle_cutter(f, r, deg=True)
            db.pline = ax_phase.scatter(
                _f, _r, label=label, color=color, marker=marker, **kwargs
            )

        midpt = (np.max(f) - np.min(f)) / 2
        if np.count_nonzero(f < midpt) < 0.65 * len(f):
            db.scale_log = False
        else:
            db.scale_log = True

        db.maxF_Hz = max(f)
        db.minF_Hz = min(f)
        select = f > 0
        db.minF_HzNZ = min(f[select])

        db.min_res_err_lim = np.min(abs(r)) / 1.2
        db.max_res_err_lim = np.max(abs(r)) * 1.2
        db.median_res_lim = np.median(abs(r))
        fit_list.append(db)
        return db

    return plot


def plot_augment_zp(self, ax):
    def plot(fitter):
        BWs = []
        ZPKrep = ZPKrep2Sf(fitter.ZPKrep)
        p = ZPKrep.poles.fullplane
        F = p.imag
        BW = p.real
        select = BW < 0
        ax.errorbar(
            F[select],
            -BW[select],
            xerr=-BW[select],
            color=self.poleID_color,
            lw=0.5,
            alpha=0.5,
            marker="None",
            ls="None",
        )
        ax.scatter(
            F[select],
            -BW[select],
            s=10,
            facecolor=self.poleID_color,
            linewidths=0.5,
            # edgecolors = 'black',
        )
        BWs.extend(-BW[select])
        ax.errorbar(
            F[~select],
            BW[~select],
            xerr=BW[~select],
            color=self.poleOD_color,
            lw=0.5,
            alpha=0.5,
            marker="None",
            ls="None",
        )
        ax.scatter(
            F[~select],
            BW[~select],
            s=10,
            facecolor=self.poleOD_color,
            linewidths=0.5,
            # edgecolors = 'black',
        )
        BWs.extend(BW[~select])

        z = ZPKrep.zeros.fullplane
        F = z.imag
        BW = z.real
        select = BW < 0
        ax.errorbar(
            F[select],
            -BW[select],
            xerr=-BW[select],
            color=self.zeroID_color,
            lw=0.5,
            alpha=0.5,
            marker="None",
            ls="None",
        )
        ax.scatter(
            F[select],
            -BW[select],
            s=10,
            facecolor=self.zeroID_color,
            linewidths=0.5,
            # edgecolors = 'black',
            marker="x",
        )
        BWs.extend(-BW[select])
        ax.errorbar(
            F[~select],
            BW[~select],
            xerr=BW[~select],
            color=self.zeroOD_color,
            lw=0.5,
            alpha=0.5,
            marker="None",
            ls="None",
        )
        ax.scatter(
            F[~select],
            BW[~select],
            s=10,
            facecolor=self.zeroOD_color,
            linewidths=0.5,
            # edgecolors = 'black',
            marker="x",
        )
        BWs.extend(BW[~select])
        BWs = np.asarray(BWs)

        if len(z) > 0 or len(p) > 0:
            select = BWs > 0
            BWs = BWs[select]
            ax.set_ylim(np.max(BWs), np.min(BWs))
            ax.set_yscale("log_zoom")
        return

    return plot
